#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import datetime
import unittest

import pytest
from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.datetime.min_max_datetime import MinMaxDatetime
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.stream_slicers.datetime_stream_slicer import DatetimeStreamSlicer

datetime_format = "%Y-%m-%d"
FAKE_NOW = datetime.datetime(2022, 1, 1, tzinfo=datetime.timezone.utc)

config = {"start_date": "2021-01-01"}
end_date_now = InterpolatedString(
    "{{ today_utc() }}",
)
cursor_value = InterpolatedString("{{ stream_state['date'] }}")
timezone = datetime.timezone.utc


@pytest.fixture()
def mock_datetime_now(monkeypatch):
    datetime_mock = unittest.mock.MagicMock(wraps=datetime.datetime)
    datetime_mock.now.return_value = FAKE_NOW
    monkeypatch.setattr(datetime, "datetime", datetime_mock)


@pytest.mark.parametrize(
    "test_name, stream_state, start, end, step, cursor, lookback_window, expected_slices",
    [
        (
            "test_1_day",
            None,
            MinMaxDatetime("{{ config['start_date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            "1d",
            cursor_value,
            None,
            [
                {"start_date": "2021-01-01", "end_date": "2021-01-01"},
                {"start_date": "2021-01-02", "end_date": "2021-01-02"},
                {"start_date": "2021-01-03", "end_date": "2021-01-03"},
                {"start_date": "2021-01-04", "end_date": "2021-01-04"},
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
                {"start_date": "2021-01-06", "end_date": "2021-01-06"},
                {"start_date": "2021-01-07", "end_date": "2021-01-07"},
                {"start_date": "2021-01-08", "end_date": "2021-01-08"},
                {"start_date": "2021-01-09", "end_date": "2021-01-09"},
                {"start_date": "2021-01-10", "end_date": "2021-01-10"},
            ],
        ),
        (
            "test_2_day",
            None,
            MinMaxDatetime("{{ config['start_date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            "2d",
            cursor_value,
            None,
            [
                {"start_date": "2021-01-01", "end_date": "2021-01-02"},
                {"start_date": "2021-01-03", "end_date": "2021-01-04"},
                {"start_date": "2021-01-05", "end_date": "2021-01-06"},
                {"start_date": "2021-01-07", "end_date": "2021-01-08"},
                {"start_date": "2021-01-09", "end_date": "2021-01-10"},
            ],
        ),
        (
            "test_from_stream_state",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ stream_state['date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            "1d",
            cursor_value,
            None,
            [
                # FIXME: should this include 2021-01-05?
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
                {"start_date": "2021-01-06", "end_date": "2021-01-06"},
                {"start_date": "2021-01-07", "end_date": "2021-01-07"},
                {"start_date": "2021-01-08", "end_date": "2021-01-08"},
                {"start_date": "2021-01-09", "end_date": "2021-01-09"},
                {"start_date": "2021-01-10", "end_date": "2021-01-10"},
            ],
        ),
        (
            "test_12_day",
            None,
            MinMaxDatetime("{{ config['start_date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            "12d",
            cursor_value,
            None,
            [
                {"start_date": "2021-01-01", "end_date": "2021-01-10"},
            ],
        ),
        (
            "test_end_date_greater_than_now",
            None,
            MinMaxDatetime("2021-12-28", datetime_format=datetime_format),
            MinMaxDatetime(f"{(FAKE_NOW + datetime.timedelta(days=1)).strftime(datetime_format)}", datetime_format=datetime_format),
            "1d",
            cursor_value,
            None,
            [
                {"start_date": "2021-12-28", "end_date": "2021-12-28"},
                {"start_date": "2021-12-29", "end_date": "2021-12-29"},
                {"start_date": "2021-12-30", "end_date": "2021-12-30"},
                {"start_date": "2021-12-31", "end_date": "2021-12-31"},
                {"start_date": "2022-01-01", "end_date": "2022-01-01"},
            ],
        ),
        (
            "test_start_date_greater_than_end_date",
            {"date": "2021-01-05"},
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            MinMaxDatetime("{{ stream_state['date'] }}", datetime_format=datetime_format),
            "1d",
            cursor_value,
            None,
            [
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
            ],
        ),
        (
            "test_cursor_date_greater_than_start_date",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ stream_state['date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            "1d",
            InterpolatedString("{{ stream_state['date'] }}"),
            None,
            [
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
                {"start_date": "2021-01-06", "end_date": "2021-01-06"},
                {"start_date": "2021-01-07", "end_date": "2021-01-07"},
                {"start_date": "2021-01-08", "end_date": "2021-01-08"},
                {"start_date": "2021-01-09", "end_date": "2021-01-09"},
                {"start_date": "2021-01-10", "end_date": "2021-01-10"},
            ],
        ),
        (
            "test_start_date_less_than_min_date",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ config['start_date'] }}", min_datetime="{{ stream_state['date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", datetime_format=datetime_format),
            "1d",
            InterpolatedString("{{ stream_state['date'] }}"),
            None,
            [
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
                {"start_date": "2021-01-06", "end_date": "2021-01-06"},
                {"start_date": "2021-01-07", "end_date": "2021-01-07"},
                {"start_date": "2021-01-08", "end_date": "2021-01-08"},
                {"start_date": "2021-01-09", "end_date": "2021-01-09"},
                {"start_date": "2021-01-10", "end_date": "2021-01-10"},
            ],
        ),
        (
            "test_end_date_greater_than_max_date",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ config['start_date'] }}", datetime_format=datetime_format),
            MinMaxDatetime("2021-01-10", max_datetime="{{ stream_state['date'] }}", datetime_format=datetime_format),
            "1d",
            None,
            None,
            [
                {"start_date": "2021-01-01", "end_date": "2021-01-01"},
                {"start_date": "2021-01-02", "end_date": "2021-01-02"},
                {"start_date": "2021-01-03", "end_date": "2021-01-03"},
                {"start_date": "2021-01-04", "end_date": "2021-01-04"},
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
            ],
        ),
        (
            "test_start_end_min_max_inherits_datetime_format_from_stream_slicer",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ config['start_date'] }}"),
            MinMaxDatetime("2021-01-10", max_datetime="{{ stream_state['date'] }}"),
            "1d",
            None,
            None,
            [
                {"start_date": "2021-01-01", "end_date": "2021-01-01"},
                {"start_date": "2021-01-02", "end_date": "2021-01-02"},
                {"start_date": "2021-01-03", "end_date": "2021-01-03"},
                {"start_date": "2021-01-04", "end_date": "2021-01-04"},
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
            ],
        ),
        (
            "test_with_lookback_window_from_start_date",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ config['start_date'] }}"),
            MinMaxDatetime("2021-01-10", max_datetime="{{ stream_state['date'] }}"),
            "1d",
            None,
            "3d",
            [
                {"start_date": "2020-12-29", "end_date": "2020-12-29"},
                {"start_date": "2020-12-30", "end_date": "2020-12-30"},
                {"start_date": "2020-12-31", "end_date": "2020-12-31"},
                {"start_date": "2021-01-01", "end_date": "2021-01-01"},
                {"start_date": "2021-01-02", "end_date": "2021-01-02"},
                {"start_date": "2021-01-03", "end_date": "2021-01-03"},
                {"start_date": "2021-01-04", "end_date": "2021-01-04"},
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
            ],
        ),
        (
            "test_with_lookback_window_defaults_to_0d",
            {"date": "2021-01-05"},
            MinMaxDatetime("{{ config['start_date'] }}"),
            MinMaxDatetime("2021-01-10", max_datetime="{{ stream_state['date'] }}"),
            "1d",
            None,
            "{{ config['does_not_exist'] }}",
            [
                {"start_date": "2021-01-01", "end_date": "2021-01-01"},
                {"start_date": "2021-01-02", "end_date": "2021-01-02"},
                {"start_date": "2021-01-03", "end_date": "2021-01-03"},
                {"start_date": "2021-01-04", "end_date": "2021-01-04"},
                {"start_date": "2021-01-05", "end_date": "2021-01-05"},
            ],
        ),
    ],
)
def test_stream_slices(mock_datetime_now, test_name, stream_state, start, end, cursor, step, lookback_window, expected_slices):
    lookback_window = InterpolatedString(lookback_window) if lookback_window else None
    slicer = DatetimeStreamSlicer(
        start_datetime=start,
        end_datetime=end,
        step=step,
        cursor_value=cursor,
        datetime_format=datetime_format,
        lookback_window=lookback_window,
        config=config,
    )
    stream_slices = slicer.stream_slices(SyncMode.incremental, stream_state)

    assert expected_slices == stream_slices


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import json
import os
import random
import shutil
import string
from pathlib import Path
from typing import Any, List, Mapping, Tuple

import pytest
from airbyte_cdk.sources.streams.files.formats.csv_parser import CsvParser

from .abstract_parser_tests import AbstractParserTests, memory_limit
from .conftest import TMP_FOLDER

TEST_FILES_DIR = Path(__file__).resolve().parent.joinpath("test_files/")

# All possible CSV data types
CSV_TYPES = {
    # logical_type: (json_type, csv_types, convert_function)
    # standard types
    "string": ("string", ["string"], None),
    "boolean": ("boolean", ["boolean"], None),
    "number": ("number", ["number"], None),
    "integer": ("integer", ["integer"], None),
}


def _generate_value(typ: str) -> Any:
    if typ == "string":
        if AbstractParserTests._generate_value("boolean"):
            return None
        random_length = random.randint(0, 512)
        return "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(random_length))
    return AbstractParserTests._generate_value(typ)


def _generate_row(types: List[str]) -> List[Any]:
    """Generates random values with request types"""
    row = []
    for needed_type in types:
        for json_type in CSV_TYPES:
            if json_type == needed_type:
                value = _generate_value(needed_type)
                if value is None:
                    value = ""
                row.append(str(value))
                break
    return row


def generate_csv_file(filename: str, columns: Mapping[str, str], num_rows: int, delimiter: str) -> str:
    """Generates  a random CSV data and save it to a tmp file"""
    header_line = delimiter.join(columns.keys())
    types = list(columns.values()) if num_rows else []
    with open(filename, "w") as f:
        f.write(header_line + "\n")
        for _ in range(num_rows):
            f.write(delimiter.join(_generate_row(types)) + "\n")
    return filename


def generate_big_file(filepath: str, size_in_gigabytes: float, columns_number: int, template_file: str = None) -> Tuple[dict, float]:
    temp_files = [filepath + ".1", filepath + ".2"]
    if template_file:
        shutil.copyfile(template_file, filepath)
        schema = None
    else:
        schema = {f"column {i}": random.choice(["integer", "string", "boolean", "number"]) for i in range(columns_number)}
        generate_csv_file(filepath, schema, 456, ",")

    skip_headers = False
    with open(filepath, "r") as f:
        with open(temp_files[0], "w") as tf:
            for line in f:
                if not skip_headers:
                    skip_headers = True
                    continue
                tf.write(str(line))

    with open(filepath, "ab") as f:
        while True:
            file_size = os.stat(filepath).st_size / (1024**3)
            if file_size > size_in_gigabytes:
                break
            with open(temp_files[0], "rb") as tf:  # type: ignore[assignment]
                with open(temp_files[1], "wb") as tf2:
                    buf = tf.read(50 * 1024**2)  # by 50Mb
                    if buf:
                        f.write(buf)  # type: ignore[arg-type]
                        tf2.write(buf)  # type: ignore[arg-type]
            temp_files.append(temp_files.pop(0))
    # remove temp files
    for temp_file in temp_files:
        if os.path.exists(temp_file):
            os.remove(temp_file)
    return schema, file_size


class TestCsvParser(AbstractParserTests):
    record_types = CSV_TYPES
    filetype = "csv"

    @classmethod
    def cases(cls) -> Mapping[str, Any]:
        return {
            "basic_normal_test": {
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv"},
                    master_schema={
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_1.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
            "infer_schema_turned_off": {
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv", "infer_datatypes": False},
                    master_schema={
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_1.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "string",
                    "name": "string",
                    "valid": "string",
                    "code": "string",
                    "degrees": "string",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
            "custom_csv_parameters": {
                # tests custom CSV parameters (odd delimiter, quote_char, escape_char & newlines in values in the file)
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv", "delimiter": "^", "quote_char": "|", "escape_char": "!", "newlines_in_values": True},
                    master_schema={
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_2_params.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
            "encoding_Big5": {
                # tests encoding: Big5
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv", "encoding": "big5"}, master_schema={"id": "integer", "name": "string", "valid": "boolean"}
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_3_enc_Big5.csv"),
                "num_records": 8,
                "inferred_schema": {"id": "integer", "name": "string", "valid": "boolean"},
                "line_checks": {
                    3: {
                        "id": 3,
                        "name": "變形金剛，偽裝的機器人",
                        "valid": False,
                    }
                },
                "fails": [],
            },
            "encoding_Arabic_(Windows 1256)": {
                # tests encoding: Arabic (Windows 1256)
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv", "encoding": "windows-1256"},
                    master_schema={"id": "integer", "notes": "string", "valid": "boolean"},
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_4_enc_Arabic.csv"),
                "num_records": 2,
                "inferred_schema": {"id": "integer", "notes": "string", "valid": "boolean"},
                "line_checks": {
                    1: {
                        "id": 1,
                        "notes": "البايت الجوي هو الأفضل",
                        "valid": False,
                    }
                },
                "fails": [],
            },
            "compression_gzip": {
                # tests compression: gzip
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv"},
                    master_schema={
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_5.csv.gz"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {
                    7: {
                        "id": 7,
                        "name": "xZhh1Kyl",
                        "valid": False,
                        "code": 10,
                        "degrees": -9.2,
                        "birthday": "2021-07-14",
                        "last_seen": "2021-07-14 15:30:09.225145",
                    }
                },
                "fails": [],
            },
            "compression_bz2": {
                # tests compression: bz2
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv"},
                    master_schema={
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_7_bz2.csv.bz2"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {
                    7: {
                        "id": 7,
                        "name": "xZhh1Kyl",
                        "valid": False,
                        "code": 10,
                        "degrees": -9.2,
                        "birthday": "2021-07-14",
                        "last_seen": "2021-07-14 15:30:09.225145",
                    }
                },
                "fails": [],
            },
            "extra_columns_in_master_schema": {
                # tests extra columns in master schema
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv"},
                    master_schema={
                        "EXTRA_COLUMN_1": "boolean",
                        "EXTRA_COLUMN_2": "number",
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_1.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
            "missing_columns_in_master_schema": {
                # tests missing columns in master schema
                # TODO: maybe this should fail read_records, but it does pick up all the columns from file despite missing from master schema
                "AbstractFileParser": CsvParser(format={"filetype": "csv"}, master_schema={"id": "integer", "name": "string"}),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_1.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
            "empty_csv_file": {
                # tests empty file, SHOULD FAIL INFER & STREAM RECORDS
                "AbstractFileParser": CsvParser(format={"filetype": "csv"}, master_schema={}),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_6_empty.csv"),
                "num_records": 0,
                "inferred_schema": {},
                "line_checks": {},
                "fails": ["test_get_inferred_schema", "test_stream_records"],
            },
            "empty_advanced_options": {
                "AbstractFileParser": CsvParser(
                    format={"filetype": "csv", "advanced_options": ""},
                    master_schema={
                        "id": "integer",
                        "name": "string",
                        "valid": "boolean",
                        "code": "integer",
                        "degrees": "number",
                        "birthday": "string",
                        "last_seen": "string",
                    },
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_1.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
            "no_header_csv_file": {
                # no header test
                "AbstractFileParser": CsvParser(
                    format={
                        "filetype": "csv",
                        "advanced_options": json.dumps(
                            {"column_names": ["id", "name", "valid", "code", "degrees", "birthday", "last_seen"]}
                        ),
                    },
                    master_schema={},
                ),
                "filepath": os.path.join(TEST_FILES_DIR, "csv/test_file_8_no_header.csv"),
                "num_records": 8,
                "inferred_schema": {
                    "id": "integer",
                    "name": "string",
                    "valid": "boolean",
                    "code": "integer",
                    "degrees": "number",
                    "birthday": "string",
                    "last_seen": "string",
                },
                "line_checks": {},
                "fails": [],
            },
        }

    @memory_limit(20)
    @pytest.mark.order(1)
    def test_big_file(self) -> None:
        """tests a big csv file (>= 1.5G records)"""
        filepath = os.path.join(TMP_FOLDER, "big_csv_file." + self.filetype)
        schema, file_size = generate_big_file(filepath, 0.1, 123)
        expected_count = sum(1 for _ in open(filepath)) - 1
        self.logger.info(f"generated file {filepath} with size {file_size}Gb, lines: {expected_count}")
        for _ in range(3):
            parser = CsvParser(
                format={"filetype": self.filetype, "block_size": 5 * 1024**2},
                master_schema=schema,
            )
            expected_file = open(filepath, "r")
            # skip the first header line
            next(expected_file)
            read_count = 0
            with self._open_and_handle_decompress(filepath, self._get_readmode({"AbstractFileParser": parser})) as f:
                for record in parser.stream_records(f):
                    record_line = ",".join("" if v is None else str(v) for v in record.values())
                    expected_line = next(expected_file).strip("\n")
                    assert record_line == expected_line
                    read_count += 1
            assert read_count == expected_count
            expected_file.close()



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString

config = {"start": 1234}
kwargs = {"next_page_token": {"next_page_url": "https://airbyte.io"}}


def test_value_is_static():
    static_value = "a_static_value"
    interpolated_string = InterpolatedString(static_value)

    evaluated_string = interpolated_string.eval(config, **kwargs)

    assert evaluated_string == static_value


def test_value_from_config():
    string = "{{ config['start'] }}"
    interpolated_string = InterpolatedString(string)

    evaluated_string = interpolated_string.eval(config, **kwargs)

    assert evaluated_string == config["start"]


def test_value_from_kwargs():
    string = "{{ next_page_token['next_page_url'] }}"
    interpolated_string = InterpolatedString(string)

    evaluated_string = interpolated_string.eval(config, **kwargs)

    assert evaluated_string == "https://airbyte.io"


def test_default_value():
    static_value = "{{ config['end'] }}"
    default = 5678
    interpolated_string = InterpolatedString(static_value, default)

    evaluated_string = interpolated_string.eval(config, **kwargs)

    assert evaluated_string == default


def test_interpolated_default_value():
    static_value = "{{ config['end'] }}"
    interpolated_string = InterpolatedString(static_value, "{{ config['start'] }}")

    evaluated_string = interpolated_string.eval(config, **kwargs)

    assert evaluated_string == config["start"]



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, Iterable, Mapping

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer


class SingleSlice(StreamSlicer):
    def __init__(self, **kwargs):
        pass

    def stream_slices(self, sync_mode: SyncMode, stream_state: Mapping[str, Any]) -> Iterable[Mapping[str, Any]]:
        return [dict()]



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Union

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.retrievers.retriever import Retriever
from airbyte_cdk.sources.declarative.schema.schema_loader import SchemaLoader
from airbyte_cdk.sources.declarative.transformations import RecordTransformation
from airbyte_cdk.sources.declarative.types import Config, StreamSlice
from airbyte_cdk.sources.streams.core import Stream


class DeclarativeStream(Stream):
    """
    DeclarativeStream is a Stream that delegates most of its logic to its schema_load and retriever
    """

    def __init__(
        self,
        name: str,
        primary_key,
        schema_loader: SchemaLoader,
        retriever: Retriever,
        config: Config,
        cursor_field: Optional[List[str]] = None,
        transformations: List[RecordTransformation] = None,
        checkpoint_interval: Optional[int] = None,
    ):
        """

        :param name: stream name
        :param primary_key: the primary key of the stream
        :param schema_loader:
        :param retriever:
        :param cursor_field:
        :param transformations: A list of transformations to be applied to each output record in the stream. Transformations are applied
        in the order in which they are defined.
        """
        self._name = name
        self._config = config
        self._primary_key = primary_key
        self._cursor_field = cursor_field or []
        self._schema_loader = schema_loader
        self._retriever = retriever
        self._transformations = transformations or []
        self._checkpoint_interval = checkpoint_interval

    @property
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        return self._primary_key

    @property
    def name(self) -> str:
        """
        :return: Stream name. By default this is the implementing class name, but it can be overridden as needed.
        """
        return self._name

    @property
    def state_checkpoint_interval(self) -> Optional[int]:
        """
        Decides how often to checkpoint state (i.e: emit a STATE message). E.g: if this returns a value of 100, then state is persisted after reading
        100 records, then 200, 300, etc.. A good default value is 1000 although your mileage may vary depending on the underlying data source.

        Checkpointing a stream avoids re-reading records in the case a sync is failed or cancelled.

        return None if state should not be checkpointed e.g: because records returned from the underlying data source are not returned in
        ascending order with respect to the cursor field. This can happen if the source does not support reading records in ascending order of
        created_at date (or whatever the cursor is). In those cases, state must only be saved once the full stream has been read.
        """
        return self._checkpoint_interval

    @property
    def state(self) -> MutableMapping[str, Any]:
        return self._retriever.state

    @state.setter
    def state(self, value: MutableMapping[str, Any]):
        """State setter, accept state serialized by state getter."""
        self._retriever.state = value

    def get_updated_state(self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]):
        return self.state

    @property
    def cursor_field(self) -> Union[str, List[str]]:
        """
        Override to return the default cursor field used by this stream e.g: an API entity might always use created_at as the cursor field.
        :return: The name of the field used as a cursor. If the cursor is nested, return an array consisting of the path to the cursor.
        """
        return self._cursor_field

    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: List[str] = None,
        stream_slice: Mapping[str, Any] = None,
        stream_state: Mapping[str, Any] = None,
    ) -> Iterable[Mapping[str, Any]]:
        for record in self._retriever.read_records(sync_mode, cursor_field, stream_slice, stream_state):
            yield self._apply_transformations(record, self._config, stream_slice)

    def _apply_transformations(self, record: Mapping[str, Any], config: Config, stream_slice: StreamSlice):
        output_record = record
        for transformation in self._transformations:
            output_record = transformation.transform(record, config=config, stream_state=self.state, stream_slice=stream_slice)

        return output_record

    def get_json_schema(self) -> Mapping[str, Any]:
        """
        :return: A dict of the JSON schema representing this stream.

        The default implementation of this method looks for a JSONSchema file with the same name as this stream's "name" property.
        Override as needed.
        """
        # TODO show an example of using pydantic to define the JSON schema, or reading an OpenAPI spec
        return self._schema_loader.get_json_schema()

    def stream_slices(
        self, *, sync_mode: SyncMode, cursor_field: List[str] = None, stream_state: Mapping[str, Any] = None
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        """
        Override to define the slices for this stream. See the stream slicing section of the docs for more information.

        :param sync_mode:
        :param cursor_field:
        :param stream_state:
        :return:
        """
        # this is not passing the cursor field because it is known at init time
        return self._retriever.stream_slices(sync_mode=sync_mode, stream_state=stream_state)



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import List, Optional

from pydantic import BaseModel, Field


class ParquetFormat(BaseModel):
    'This connector utilises <a href="https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetFile.html" target="_blank">PyArrow (Apache Arrow)</a> for Parquet parsing.'

    class Config:
        title = "Parquet"

    filetype: str = Field("parquet", const=True)

    columns: Optional[List[str]] = Field(
        default=None,
        description="If you only want to sync a subset of the columns from the file(s), add the columns you want here as a comma-delimited"
        " list. Leave it empty to sync all columns.",
        title="Selected Columns",
        order=0,
    )
    batch_size: int = Field(
        title="Record batch size",
        order=1,
        default=64 * 1024,  # 64K records
        description="Maximum number of records per batch read from the input files. "
        "Batches may be smaller if there aren’t enough rows in the file. "
        "This option can help avoid out-of-memory errors if your data is particularly wide.",
    )
    buffer_size: int = Field(
        default=2,
        description="Perform read buffering when deserializing individual column chunks. "
        "By default every group column will be loaded fully to memory. "
        "This option can help avoid out-of-memory errors if your data is particularly wide.",
    )



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import logging
import time
from typing import Any, Mapping, Tuple

import airbyte_cdk.sources.streams.files.utils as utils
import pyarrow as pa
import pytest

logger = logging.getLogger("airbyte")


def external_process_failure():
    try:
        1 / 0
    except Exception as e:
        return (None, e)


def external_process_success(dummy_arg):
    return (dummy_arg, None)


def external_process_timeout():
    time.sleep(6)


class TestFilesUtils:
    def test_run_in_external_process_failure(self):
        with pytest.raises(ZeroDivisionError):
            utils.run_in_external_process(external_process_failure, 2, 4, logger, [])

    def test_run_in_external_process_success(self):
        arg_to_pass = "7b372d17-c3d1-4e44-95ed-b8edb3790160"
        return_val = utils.run_in_external_process(external_process_success, 2, 4, logger, [arg_to_pass])
        assert return_val == arg_to_pass

    def test_run_in_external_process_timeout(self):
        with pytest.raises(TimeoutError):
            utils.run_in_external_process(external_process_timeout, 1, 2, logger, [])

    def test_get_value_or_json_if_empty_string(self):
        assert utils.get_value_or_json_if_empty_string("   abcd efgh   ") == "abcd efgh"
        assert utils.get_value_or_json_if_empty_string("") == "{}"

    @pytest.mark.parametrize(  # testing all datatypes as laid out here: https://json-schema.org/understanding-json-schema/reference/type.html
        "input_json_type, output_pyarrow_type",
        [
            ("string", pa.large_string()),
            ("number", pa.float64()),
            ("integer", pa.int64()),
            ("object", pa.large_string()),
            ("array", pa.large_string()),
            ("boolean", pa.bool_()),
            ("null", pa.large_string()),
        ],
    )
    def test_json_type_to_pyarrow_type(self, input_json_type: str, output_pyarrow_type: Any) -> None:
        # Json -> PyArrow direction
        logger.info(f"asserting that JSON type '{input_json_type}' converts to PyArrow type '{output_pyarrow_type}'...")
        assert utils.json_type_to_pyarrow_type(input_json_type, logger) == output_pyarrow_type

    @pytest.mark.parametrize(  # testing all datatypes as laid out here: https://arrow.apache.org/docs/python/api/datatypes.html
        "input_pyarrow_types, output_json_type",
        [
            ((pa.null(),), "string"),  # null type
            ((pa.bool_(),), "boolean"),  # boolean type
            (
                (pa.int8(), pa.int16(), pa.int32(), pa.int64(), pa.uint8(), pa.uint16(), pa.uint32(), pa.uint64()),
                "integer",
            ),  # integer types
            ((pa.float16(), pa.float32(), pa.float64(), pa.decimal128(5, 10), pa.decimal256(3, 8)), "number"),  # number types
            ((pa.time32("s"), pa.time64("ns"), pa.timestamp("ms"), pa.date32(), pa.date64()), "string"),  # temporal types
            ((pa.binary(), pa.large_binary()), "string"),  # binary types
            ((pa.string(), pa.utf8(), pa.large_string(), pa.large_utf8()), "string"),  # string types
            ((pa.list_(pa.string()), pa.large_list(pa.timestamp("us"))), "string"),  # array types
            ((pa.map_(pa.string(), pa.float32()), pa.dictionary(pa.int16(), pa.list_(pa.string()))), "string"),  # object types
        ],
    )
    def test_json_type_to_pyarrow_type_reverse(self, input_pyarrow_types: Tuple[Any], output_json_type: str) -> None:
        # PyArrow -> Json direction (reverse=True)
        for typ in input_pyarrow_types:
            logger.info(f"asserting that PyArrow type '{typ}' converts to JSON type '{output_json_type}'...")
            assert utils.json_type_to_pyarrow_type(typ, logger, reverse=True) == output_json_type

    @pytest.mark.parametrize(  # if expecting fail, put pyarrow_schema as None
        "json_schema, pyarrow_schema",
        [
            (
                {"a": "string", "b": "number", "c": "integer", "d": "object", "e": "array", "f": "boolean", "g": "null"},
                {
                    "a": pa.large_string(),
                    "b": pa.float64(),
                    "c": pa.int64(),
                    "d": pa.large_string(),
                    "e": pa.large_string(),
                    "f": pa.bool_(),
                    "g": pa.large_string(),
                },
            ),
            ({"single_column": "object"}, {"single_column": pa.large_string()}),
            ({}, {}),
            ({"a": "NOT A REAL TYPE", "b": "another fake type"}, {"a": pa.large_string(), "b": pa.large_string()}),
            (["string", "object"], None),  # bad input type
        ],
    )
    def test_json_schema_to_pyarrow_schema(self, json_schema: Mapping[str, Any], pyarrow_schema: Mapping[str, Any]) -> None:
        # Json -> PyArrow direction
        if pyarrow_schema is not None:
            assert utils.json_schema_to_pyarrow_schema(json_schema, logger) == pyarrow_schema
        else:
            with pytest.raises(Exception) as e_info:
                utils.json_schema_to_pyarrow_schema(json_schema, logger)
                logger.debug(str(e_info))

    @pytest.mark.parametrize(  # if expecting fail, put json_schema as None
        "pyarrow_schema, json_schema",
        [
            (
                {
                    "a": pa.utf8(),
                    "b": pa.float16(),
                    "c": pa.uint32(),
                    "d": pa.map_(pa.string(), pa.float32()),
                    "e": pa.bool_(),
                    "f": pa.date64(),
                },
                {"a": "string", "b": "number", "c": "integer", "d": "string", "e": "boolean", "f": "string"},
            ),
            ({"single_column": pa.int32()}, {"single_column": "integer"}),
            ({}, {}),
            ({"a": "NOT A REAL TYPE", "b": "another fake type"}, {"a": "string", "b": "string"}),
            (["string", "object"], None),  # bad input type
        ],
    )
    def test_json_schema_to_pyarrow_schema_reverse(self, pyarrow_schema: Mapping[str, Any], json_schema: Mapping[str, Any]) -> None:
        # PyArrow -> Json direction (reverse=True)
        if json_schema is not None:
            assert utils.json_schema_to_pyarrow_schema(pyarrow_schema, logger, reverse=True) == json_schema
        else:
            with pytest.raises(Exception) as e_info:
                utils.json_schema_to_pyarrow_schema(pyarrow_schema, logger, reverse=True)
                logger.debug(str(e_info))



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from unittest.mock import MagicMock

import airbyte_cdk.sources.declarative.requesters.error_handlers.response_status as response_status
import pytest
import requests
from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_action import ResponseAction
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_status import ResponseStatus
from airbyte_cdk.sources.declarative.requesters.requester import HttpMethod
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever

primary_key = "pk"
records = [{"id": 1}, {"id": 2}]


def test_simple_retriever():
    requester = MagicMock()
    request_params = {"param": "value"}
    requester.request_params.return_value = request_params

    paginator = MagicMock()
    next_page_token = {"cursor": "cursor_value"}
    paginator.next_page_token.return_value = next_page_token

    record_selector = MagicMock()
    record_selector.select_records.return_value = records

    iterator = MagicMock()
    stream_slices = [{"date": "2022-01-01"}, {"date": "2022-01-02"}]
    iterator.stream_slices.return_value = stream_slices

    response = requests.Response()

    state = MagicMock()
    underlying_state = {"date": "2021-01-01"}
    state.get_stream_state.return_value = underlying_state

    url_base = "https://airbyte.io"
    requester.get_url_base.return_value = url_base
    path = "/v1"
    requester.get_path.return_value = path
    http_method = HttpMethod.GET
    requester.get_method.return_value = http_method
    backoff_time = 60
    should_retry = ResponseStatus.retry(backoff_time)
    requester.should_retry.return_value = should_retry
    request_body_data = {"body": "data"}
    requester.request_body_data.return_value = request_body_data
    request_body_json = {"body": "json"}
    requester.request_body_json.return_value = request_body_json
    request_kwargs = {"kwarg": "value"}
    requester.request_kwargs.return_value = request_kwargs
    cache_filename = "cache"
    requester.cache_filename = cache_filename
    use_cache = True
    requester.use_cache = use_cache

    retriever = SimpleRetriever(
        "stream_name",
        primary_key,
        requester=requester,
        paginator=paginator,
        record_selector=record_selector,
        stream_slicer=iterator,
        state=state,
    )

    # hack because we clone the state...
    retriever._state = state

    assert retriever.primary_key == primary_key
    assert retriever.url_base == url_base
    assert retriever.path() == path
    assert retriever.state == underlying_state
    assert retriever.next_page_token(response) == next_page_token
    assert retriever.request_params(None, None, None) == request_params
    assert retriever.stream_slices(sync_mode=SyncMode.incremental) == stream_slices

    assert retriever._last_response is None
    assert retriever._last_records is None
    assert retriever.parse_response(response, stream_state=None) == records
    assert retriever._last_response == response
    assert retriever._last_records == records

    assert retriever.http_method == "GET"
    assert not retriever.raise_on_http_errors
    assert retriever.should_retry(requests.Response())
    assert retriever.backoff_time(requests.Response()) == backoff_time
    assert retriever.request_body_data(None, None, None) == request_body_data
    assert retriever.request_body_json(None, None, None) == request_body_json
    assert retriever.request_kwargs(None, None, None) == request_kwargs
    assert retriever.cache_filename == cache_filename
    assert retriever.use_cache == use_cache


@pytest.mark.parametrize(
    "test_name, requester_response, expected_should_retry, expected_backoff_time",
    [
        ("test_should_retry_fail", response_status.FAIL, False, None),
        ("test_should_retry_none_backoff", ResponseStatus.retry(None), True, None),
        ("test_should_retry_custom_backoff", ResponseStatus.retry(60), True, 60),
    ],
)
def test_should_retry(test_name, requester_response, expected_should_retry, expected_backoff_time):
    requester = MagicMock()
    retriever = SimpleRetriever("stream_name", primary_key, requester=requester, record_selector=MagicMock())
    requester.should_retry.return_value = requester_response
    assert retriever.should_retry(requests.Response()) == expected_should_retry
    if requester_response.action == ResponseAction.RETRY:
        assert retriever.backoff_time(requests.Response()) == expected_backoff_time


@pytest.mark.parametrize(
    "test_name, status_code, response_status, len_expected_records",
    [
        ("test_parse_response_fails_if_should_retry_is_fail", 404, response_status.FAIL, None),
        ("test_parse_response_succeeds_if_should_retry_is_ok", 200, response_status.SUCCESS, 1),
        ("test_parse_response_succeeds_if_should_retry_is_ignore", 404, response_status.IGNORE, 0),
    ],
)
def test_parse_response(test_name, status_code, response_status, len_expected_records):
    requester = MagicMock()
    record_selector = MagicMock()
    record_selector.select_records.return_value = [{"id": 100}]
    retriever = SimpleRetriever("stream_name", primary_key, requester=requester, record_selector=record_selector)
    response = requests.Response()
    response.status_code = status_code
    requester.should_retry.return_value = response_status
    if len_expected_records is None:
        try:
            retriever.parse_response(response, stream_state={})
            assert False
        except requests.exceptions.HTTPError:
            pass
    else:
        records = retriever.parse_response(response, stream_state={})
        assert len(records) == len_expected_records


@pytest.mark.parametrize(
    "test_name, response_action, retry_in, expected_backoff_time",
    [
        ("test_backoff_retriable_request", ResponseAction.RETRY, 10, 10),
        ("test_backoff_fail_request", ResponseAction.FAIL, 10, None),
        ("test_backoff_ignore_request", ResponseAction.IGNORE, 10, None),
        ("test_backoff_success_request", ResponseAction.IGNORE, 10, None),
    ],
)
def test_backoff_time(test_name, response_action, retry_in, expected_backoff_time):
    requester = MagicMock()
    record_selector = MagicMock()
    record_selector.select_records.return_value = [{"id": 100}]
    response = requests.Response()
    retriever = SimpleRetriever("stream_name", primary_key, requester=requester, record_selector=record_selector)
    if expected_backoff_time:
        requester.should_retry.return_value = ResponseStatus(response_action, retry_in)
        actual_backoff_time = retriever.backoff_time(response)
        assert expected_backoff_time == actual_backoff_time
    else:
        try:
            retriever.backoff_time(response)
            assert False
        except ValueError:
            pass



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, List, Mapping

import pendulum
from airbyte_cdk.sources.streams.http.requests_native_auth.abstract_oauth import AbstractOauth2Authenticator


class Oauth2Authenticator(AbstractOauth2Authenticator):
    """
    Generates OAuth2.0 access tokens from an OAuth2.0 refresh token and client credentials.
    The generated access token is attached to each request via the Authorization header.
    """

    def __init__(
        self,
        token_refresh_endpoint: str,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        scopes: List[str] = None,
        token_expiry_date: pendulum.DateTime = None,
        access_token_name: str = "access_token",
        expires_in_name: str = "expires_in",
        refresh_request_body: Mapping[str, Any] = None,
    ):
        self.token_refresh_endpoint = token_refresh_endpoint
        self.client_secret = client_secret
        self.client_id = client_id
        self.refresh_token = refresh_token
        self.scopes = scopes
        self.access_token_name = access_token_name
        self.expires_in_name = expires_in_name
        self.refresh_request_body = refresh_request_body

        self.token_expiry_date = token_expiry_date or pendulum.now().subtract(days=1)
        self.access_token = None

    @property
    def token_refresh_endpoint(self) -> str:
        return self._token_refresh_endpoint

    @token_refresh_endpoint.setter
    def token_refresh_endpoint(self, value: str):
        self._token_refresh_endpoint = value

    @property
    def client_id(self) -> str:
        return self._client_id

    @client_id.setter
    def client_id(self, value: str):
        self._client_id = value

    @property
    def client_secret(self) -> str:
        return self._client_secret

    @client_secret.setter
    def client_secret(self, value: str):
        self._client_secret = value

    @property
    def refresh_token(self) -> str:
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, value: str):
        self._refresh_token = value

    @property
    def access_token_name(self) -> str:
        return self._access_token_name

    @access_token_name.setter
    def access_token_name(self, value: str):
        self._access_token_name = value

    @property
    def scopes(self) -> [str]:
        return self._scopes

    @scopes.setter
    def scopes(self, value: [str]):
        self._scopes = value

    @property
    def token_expiry_date(self) -> pendulum.DateTime:
        return self._token_expiry_date

    @token_expiry_date.setter
    def token_expiry_date(self, value: pendulum.DateTime):
        self._token_expiry_date = value

    @property
    def expires_in_name(self) -> str:
        return self._expires_in_name

    @expires_in_name.setter
    def expires_in_name(self, value):
        self._expires_in_name = value

    @property
    def refresh_request_body(self) -> Mapping[str, Any]:
        return self._refresh_request_body

    @refresh_request_body.setter
    def refresh_request_body(self, value: Mapping[str, Any]):
        self._refresh_request_body = value

    @property
    def access_token(self) -> str:
        return self._access_token

    @access_token.setter
    def access_token(self, value: str):
        self._access_token = value



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, Mapping

from airbyte_cdk.sources.declarative.requesters.interpolated_request_input_provider import InterpolatedRequestInputProvider
from airbyte_cdk.sources.declarative.requesters.request_headers.request_header_provider import RequestHeaderProvider


class InterpolatedRequestHeaderProvider(RequestHeaderProvider):
    """
    Provider that takes in a dictionary of request headers and performs string interpolation on the defined templates and static
    values based on the current state of stream being processed
    """

    def __init__(self, *, config, request_headers):
        self._interpolator = InterpolatedRequestInputProvider(config=config, request_inputs=request_headers)

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        return self._interpolator.request_inputs(stream_state, stream_slice, next_page_token)



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from pydantic import BaseModel, Field


class AvroFormat(BaseModel):
    'This connector utilises <a href="https://fastavro.readthedocs.io/en/latest/" target="_blank">fastavro</a> for Avro parsing.'

    class Config:
        title = "Avro"

    filetype: str = Field(
        "avro",
        const=True,
    )



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from __future__ import annotations

import copy
import importlib
from typing import Any, Mapping, Type, Union, get_args, get_origin, get_type_hints

from airbyte_cdk.sources.declarative.create_partial import create
from airbyte_cdk.sources.declarative.interpolation.jinja import JinjaInterpolation
from airbyte_cdk.sources.declarative.parsers.class_types_registry import CLASS_TYPES_REGISTRY
from airbyte_cdk.sources.declarative.parsers.default_implementation_registry import DEFAULT_IMPLEMENTATIONS_REGISTRY
from airbyte_cdk.sources.declarative.types import Config


class DeclarativeComponentFactory:
    def __init__(self):
        self._interpolator = JinjaInterpolation()

    def create_component(self, component_definition: Mapping[str, Any], config: Config):
        """

        :param component_definition: mapping defining the object to create. It should have at least one field: `class_name`
        :param config: Connector's config
        :return: the object to create
        """
        kwargs = copy.deepcopy(component_definition)
        if "class_name" in kwargs:
            class_name = kwargs.pop("class_name")
        elif "type" in kwargs:
            class_name = CLASS_TYPES_REGISTRY[kwargs.pop("type")]
        else:
            raise ValueError(f"Failed to create component because it has no class_name or type. Definition: {component_definition}")
        return self.build(class_name, config, **kwargs)

    def build(self, class_or_class_name: Union[str, Type], config, **kwargs):
        if isinstance(class_or_class_name, str):
            class_ = self._get_class_from_fully_qualified_class_name(class_or_class_name)
        else:
            class_ = class_or_class_name

        # create components in options before propagating them
        if "options" in kwargs:
            kwargs["options"] = {k: self._create_subcomponent(k, v, kwargs, config, class_) for k, v in kwargs["options"].items()}

        updated_kwargs = {k: self._create_subcomponent(k, v, kwargs, config, class_) for k, v in kwargs.items()}
        return create(class_, config=config, **updated_kwargs)

    @staticmethod
    def _get_class_from_fully_qualified_class_name(class_name: str):
        split = class_name.split(".")
        module = ".".join(split[:-1])
        class_name = split[-1]
        return getattr(importlib.import_module(module), class_name)

    @staticmethod
    def _merge_dicts(d1, d2):
        return {**d1, **d2}

    def _create_subcomponent(self, key, definition, kwargs, config, parent_class):
        """
        There are 5 ways to define a component.
        1. dict with "class_name" field -> create an object of type "class_name"
        2. dict with "type" field -> lookup the `CLASS_TYPES_REGISTRY` to find the type of object and create an object of that type
        3. a dict with a type that can be inferred. If the parent class's constructor has type hints, we can infer the type of the object to create by looking up the `DEFAULT_IMPLEMENTATIONS_REGISTRY` map
        4. list: loop over the list and create objects for its items
        5. anything else -> return as is
        """
        if self.is_object_definition_with_class_name(definition):
            # propagate kwargs to inner objects
            definition["options"] = self._merge_dicts(kwargs.get("options", dict()), definition.get("options", dict()))
            return self.create_component(definition, config)()
        elif self.is_object_definition_with_type(definition):
            # If type is set instead of class_name, get the class_name from the CLASS_TYPES_REGISTRY
            definition["options"] = self._merge_dicts(kwargs.get("options", dict()), definition.get("options", dict()))
            object_type = definition.pop("type")
            class_name = CLASS_TYPES_REGISTRY[object_type]
            definition["class_name"] = class_name
            return self.create_component(definition, config)()
        elif isinstance(definition, dict):
            # Try to infer object type
            expected_type = self.get_default_type(key, parent_class)
            # if there is an expected type, and it's not a builtin type, then instantiate it
            # We don't have to instantiate builtin types (eg string and dict) because definition is already going to be of that type
            if expected_type and not self._is_builtin_type(expected_type):
                definition["class_name"] = expected_type
                definition["options"] = self._merge_dicts(kwargs.get("options", dict()), definition.get("options", dict()))
                return self.create_component(definition, config)()
            else:
                return definition
        elif isinstance(definition, list):
            return [
                self._create_subcomponent(
                    key, sub, self._merge_dicts(kwargs.get("options", dict()), self._get_subcomponent_options(sub)), config, parent_class
                )
                for sub in definition
            ]
        else:
            expected_type = self.get_default_type(key, parent_class)
            if expected_type and not isinstance(definition, expected_type):
                # call __init__(definition) if definition is not a dict and is not of the expected type
                # for instance, to turn a string into an InterpolatedString
                return expected_type(definition)
            else:
                return definition

    @staticmethod
    def is_object_definition_with_class_name(definition):
        return isinstance(definition, dict) and "class_name" in definition

    @staticmethod
    def is_object_definition_with_type(definition):
        return isinstance(definition, dict) and "type" in definition

    @staticmethod
    def get_default_type(parameter_name, parent_class):
        type_hints = get_type_hints(parent_class.__init__)
        interface = type_hints.get(parameter_name)
        while True:
            origin = get_origin(interface)
            if origin:
                # Unnest types until we reach the raw type
                # List[T] -> T
                # Optional[List[T]] -> T
                args = get_args(interface)
                interface = args[0]
            else:
                break

        expected_type = DEFAULT_IMPLEMENTATIONS_REGISTRY.get(interface)

        if expected_type:
            return expected_type
        else:
            return interface

    @staticmethod
    def _get_subcomponent_options(sub: Any):
        if isinstance(sub, dict):
            return sub.get("options", {})
        else:
            return {}

    @staticmethod
    def _is_builtin_type(cls) -> bool:
        if not cls:
            return False
        return cls.__module__ == "builtins"



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, BinaryIO, Iterator, Mapping, TextIO, Union

import fastavro
from fastavro import reader

from .abstract_file_parser import AbstractFileParser

# mapping from apache avro docs: https://avro.apache.org/docs/current/spec.html#schema_complex
data_type_mapping = {
    "null": "null",
    "boolean": "boolean",
    "int": "integer",
    "long": "integer",
    "float": "number",
    "double": "number",
    "bytes": "string",
    "string": "string",
    "record": "object",
    "enum": "string",
    "array": "array",
    "map": "object",
    "fixed": "string",
}


class AvroParser(AbstractFileParser):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def is_binary(self) -> bool:
        return True

    def _parse_data_type(self, data_type_mapping: dict, avro_schema: dict) -> dict:
        """Convert data types from avro to json format
        :param data_type_mapping: mapping from avro to json data types
        :param avro_schema: schema comes with the avro file
        :return schema_dict with data types converted from avro to json standards
        """
        schema_dict = {}
        for i in avro_schema["fields"]:
            data_type = i["type"]
            # If field is nullable there will be a list of types and we need to make sure to map the whole list according to data_type_mapping
            if isinstance(data_type, list):
                schema_dict[i["name"]] = [data_type_mapping[dtype] for dtype in data_type]
            # TODO: Figure out a better way to handle nested records. Currently a nested record is returned as a string
            elif isinstance(data_type, dict):
                schema_dict[i["name"]] = "string"
            elif data_type in data_type_mapping:
                schema_dict[i["name"]] = data_type_mapping[data_type]
            else:
                raise TypeError(f"unsupported data type: {data_type} found in avro file")
        return schema_dict

    def _get_avro_schema(self, file: Union[TextIO, BinaryIO]) -> dict:
        """Extract schema for records
        :param file: file-like object (opened via StorageFile)
        :return schema extracted from the avro file
        """
        avro_reader = fastavro.reader(file)
        schema = avro_reader.writer_schema
        if not schema["type"] == "record":
            unsupported_type = schema["type"]
            raise (f"Only record based avro files are supported. Found {unsupported_type}")
        else:
            return schema

    def get_inferred_schema(self, file: Union[TextIO, BinaryIO]) -> dict:
        """Return schema
        :param file: file-like object (opened via StorageFile)
        :return: mapping of {columns:datatypes} where datatypes are JsonSchema types
        """
        avro_schema = self._get_avro_schema(file)
        schema_dict = self._parse_data_type(data_type_mapping, avro_schema)
        return schema_dict

    def stream_records(self, file: Union[TextIO, BinaryIO]) -> Iterator[Mapping[str, Any]]:
        """Stream the data using a generator
        :param file: file-like object (opened via StorageFile)
        :yield: data record as a mapping of {columns:values}
        """
        avro_reader = reader(file)
        for record in avro_reader:
            yield record



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, List, Mapping, Optional

import requests
from airbyte_cdk.sources.declarative.decoders.decoder import Decoder
from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.states.dict_state import DictState


class ConditionalPaginator:
    """
    A paginator that performs pagination by incrementing a page number and stops based on a provided stop condition.
    """

    def __init__(self, stop_condition: str, state: DictState, decoder: Decoder, config):
        self._stop_condition_interpolator = InterpolatedBoolean(stop_condition)
        self._state: DictState = state
        self._decoder = decoder
        self._config = config

    def next_page_token(self, response: requests.Response, last_records: List[Mapping[str, Any]]) -> Optional[Mapping[str, Any]]:
        decoded_response = self._decoder.decode(response)
        headers = response.headers
        should_stop = self._stop_condition_interpolator.eval(
            self._config, decoded_response=decoded_response, headers=headers, last_records=last_records
        )

        if should_stop:
            return None
        next_page = self._get_page() + 1
        self._update_page_state(next_page)
        return {"page": next_page}

    def _get_page(self):
        return self._state.get_state("page")

    def _update_page_state(self, page):
        self._state.update_state(**{"page": page})



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, List, Mapping

import requests
from airbyte_cdk.sources.declarative.extractors.http_selector import HttpSelector
from airbyte_cdk.sources.declarative.extractors.jello import JelloExtractor
from airbyte_cdk.sources.declarative.extractors.record_filter import RecordFilter
from airbyte_cdk.sources.declarative.types import Record


class RecordSelector(HttpSelector):
    """
    Responsible for translating an HTTP response into a list of records by extracting records from the response and optionally filtering
    records based on a heuristic.
    """

    def __init__(self, extractor: JelloExtractor, record_filter: RecordFilter = None):
        self._extractor = extractor
        self._record_filter = record_filter

    def select_records(
        self,
        response: requests.Response,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> List[Record]:
        all_records = self._extractor.extract_records(response)
        if self._record_filter:
            return self._record_filter.filter_records(
                all_records, stream_state=stream_state, stream_slice=stream_slice, next_page_token=next_page_token
            )
        return all_records



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import ast
from typing import Any, Iterable, List, Mapping, Union

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.declarative.interpolation.interpolated_mapping import InterpolatedMapping
from airbyte_cdk.sources.declarative.interpolation.jinja import JinjaInterpolation
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.declarative.types import Config


class ListStreamSlicer(StreamSlicer):
    """
    Stream slicer that iterates over the values of a list
    If slice_values is a string, then evaluate it as literal and assert the resulting literal is a list
    """

    def __init__(self, slice_values: Union[str, List[str]], slice_definition: Mapping[str, Any], config: Config):
        if isinstance(slice_values, str):
            slice_values = ast.literal_eval(slice_values)
        assert isinstance(slice_values, list)
        self._interpolation = InterpolatedMapping(slice_definition, JinjaInterpolation())
        self._slice_values = slice_values
        self._config = config

    def stream_slices(self, sync_mode: SyncMode, stream_state: Mapping[str, Any]) -> Iterable[Mapping[str, Any]]:
        return [self._interpolation.eval(self._config, slice_value=slice_value) for slice_value in self._slice_values]



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Mapping, Type

from airbyte_cdk.sources.declarative.checks.check_stream import CheckStream
from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.declarative.datetime.min_max_datetime import MinMaxDatetime
from airbyte_cdk.sources.declarative.declarative_stream import DeclarativeStream
from airbyte_cdk.sources.declarative.decoders.decoder import Decoder
from airbyte_cdk.sources.declarative.decoders.json_decoder import JsonDecoder
from airbyte_cdk.sources.declarative.extractors.http_selector import HttpSelector
from airbyte_cdk.sources.declarative.extractors.record_selector import RecordSelector
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.requesters.error_handlers.default_error_handler import DefaultErrorHandler
from airbyte_cdk.sources.declarative.requesters.error_handlers.error_handler import ErrorHandler
from airbyte_cdk.sources.declarative.requesters.error_handlers.http_response_filter import HttpResponseFilter
from airbyte_cdk.sources.declarative.requesters.http_requester import HttpRequester
from airbyte_cdk.sources.declarative.requesters.paginators.no_pagination import NoPagination
from airbyte_cdk.sources.declarative.requesters.paginators.paginator import Paginator
from airbyte_cdk.sources.declarative.requesters.request_options.interpolated_request_options_provider import (
    InterpolatedRequestOptionsProvider,
)
from airbyte_cdk.sources.declarative.requesters.request_options.request_options_provider import RequestOptionsProvider
from airbyte_cdk.sources.declarative.requesters.requester import Requester
from airbyte_cdk.sources.declarative.retrievers.retriever import Retriever
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever
from airbyte_cdk.sources.declarative.schema.json_schema import JsonSchema
from airbyte_cdk.sources.declarative.schema.schema_loader import SchemaLoader
from airbyte_cdk.sources.declarative.states.dict_state import DictState
from airbyte_cdk.sources.declarative.states.state import State
from airbyte_cdk.sources.declarative.stream_slicers.single_slice import SingleSlice
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.streams.core import Stream

DEFAULT_IMPLEMENTATIONS_REGISTRY: Mapping[Type, Type] = {
    Requester: HttpRequester,
    Retriever: SimpleRetriever,
    SchemaLoader: JsonSchema,
    HttpSelector: RecordSelector,
    ConnectionChecker: CheckStream,
    ErrorHandler: DefaultErrorHandler,
    Decoder: JsonDecoder,
    State: DictState,
    StreamSlicer: SingleSlice,
    RequestOptionsProvider: InterpolatedRequestOptionsProvider,
    Paginator: NoPagination,
    HttpResponseFilter: HttpResponseFilter,
    Stream: DeclarativeStream,
    MinMaxDatetime: MinMaxDatetime,
    InterpolatedString: InterpolatedString,
}



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import bz2
import gzip
import linecache
import logging
import os
import random
import sys
import tracemalloc
from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import lru_cache, wraps
from typing import Any, Callable, List, Mapping

import pytest
from airbyte_cdk.sources.streams.files import FileInfo


def memory_limit(max_memory_in_megabytes: int, print_limit: int = 20) -> Callable:
    """Runs a test function by a separate process with restricted memory"""

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: List[Any], **kwargs: Any) -> Any:
            tracemalloc.start()
            result = func(*args, **kwargs)

            # get memory usage immediately after function call, we interested in "first_size" value
            first_size, first_peak = tracemalloc.get_traced_memory()
            # get snapshot immediately just in case we will use it
            snapshot = tracemalloc.take_snapshot()

            # only if we exceeded the quota, build log_messages with traces
            first_size_in_megabytes = first_size / 1024**2
            if first_size_in_megabytes > max_memory_in_megabytes:
                log_messages: List[str] = []
                top_stats = snapshot.statistics("lineno")
                for index, stat in enumerate(top_stats[:print_limit], 1):
                    frame = stat.traceback[0]
                    filename = os.sep.join(frame.filename.split(os.sep)[-2:])
                    log_messages.append("#%s: %s:%s: %.1f KiB" % (index, filename, frame.lineno, stat.size / 1024))
                    line = linecache.getline(frame.filename, frame.lineno).strip()
                    if line:
                        log_messages.append(f"    {line}")
                traceback_log = "\n".join(log_messages)
                assert False, f"Overuse of memory, used: {first_size_in_megabytes}Mb, limit: {max_memory_in_megabytes}Mb!!\n{traceback_log}"

            return result

        return wrapper

    return decorator


def create_by_local_file(filepath: str) -> FileInfo:
    "Generates a FileInfo instance for local files"
    if not os.path.exists(filepath):
        return FileInfo(key=filepath, size=0, last_modified=datetime.now())
    return FileInfo(key=filepath, size=os.stat(filepath).st_size, last_modified=datetime.fromtimestamp(os.path.getmtime(filepath)))


class AbstractParserTests(ABC):
    """Prefixed this class with Abstract so the tests don't run here but only in the children"""

    logger = logging.getLogger("airbyte")
    record_types: Mapping[str, Any] = {}

    @classmethod
    def _generate_row(cls, types: List[str]) -> List[Any]:
        """Generates random values with request types"""
        row = []
        for needed_type in types:
            for json_type in cls.record_types:
                if json_type == needed_type:
                    row.append(cls._generate_value(needed_type))
                    break
        return row

    @classmethod
    def _generate_value(cls, typ: str) -> Any:
        if typ not in ["boolean", "integer"] and cls._generate_value("boolean"):
            # return 'None' for +- 33% of all requests
            return None

        if typ == "number":
            while True:
                int_value = cls._generate_value("integer")
                if int_value:
                    break
            return float(int_value) + random.random()
        elif typ == "integer":
            return random.randint(-sys.maxsize - 1, sys.maxsize)
            # return random.randint(0, 1000)
        elif typ == "boolean":
            return random.choice([True, False, None])
        elif typ == "string":
            random_length = random.randint(0, 10 * 1024)  # max size of bytes is 10k
            return os.urandom(random_length)
        elif typ == "timestamp":
            return datetime.now() + timedelta(seconds=random.randint(0, 7200))
        elif typ == "date":
            dt = cls._generate_value("timestamp")
            return dt.date() if dt else None
        elif typ == "time":
            dt = cls._generate_value("timestamp")
            return dt.time() if dt else None
        raise Exception(f"not supported type: {typ}")

    @classmethod
    @lru_cache(maxsize=None)
    def cached_cases(cls) -> Mapping[str, Any]:
        return cls.cases()

    @classmethod
    @abstractmethod
    def cases(cls) -> Mapping[str, Any]:
        """return a map of test cases in this structure:
        {
           "test_name": {
                "AbstractFileParser": SpecificParserClass(
                    format={format dict as expected by the parser},  # e.g. {"filetype": "csv", "encoding": "big5"}
                    master_schema={dict of actual schema of test file}  # e.g. {"id": "integer", "name": "string"}
                ),
                "filepath": "{point at path to test file}",
                "num_records": {integer value for number of rows to expect from file},
                "inferred_schema": {expected schema from parser inference},  # e.g. {"id": "integer", "name": "string"}
                "line_checks": {check specific lines are parsed correctly},  # e.g. {4: {"id": 4, "name": "bingo"} }
                "fails": []  # empty for pass. Add "test_get_inferred_schema" and/or "test_stream_records" to test fails
            },
           "test_2": etc.
        }
        note: line_checks index is 1-based to align with row numbers
        """

    def _get_readmode(self, file_info: Mapping[str, Any]) -> str:
        return "rb" if file_info["AbstractFileParser"].is_binary else "r"

    @contextmanager
    def _open_and_handle_decompress(self, filepath: str, readmode: str):
        if filepath.endswith(".gz"):
            resource = gzip.open(filepath, readmode)
        elif filepath.endswith(".bz2"):
            resource = bz2.open(filepath, readmode)
        else:
            resource = open(filepath, readmode)
        try:
            yield resource
        finally:
            resource.close()

    @memory_limit(1024)
    def test_suite_inferred_schema(self, file_info: Mapping[str, Any]) -> None:
        with self._open_and_handle_decompress(file_info["filepath"], self._get_readmode(file_info)) as f:
            if "test_get_inferred_schema" in file_info["fails"]:
                with pytest.raises(Exception) as e_info:
                    file_info["AbstractFileParser"].get_inferred_schema(f)
                    self.logger.debug(str(e_info))
            else:
                assert file_info["AbstractFileParser"].get_inferred_schema(f) == file_info["inferred_schema"]

    @memory_limit(1024)
    def test_stream_suite_records(self, file_info: Mapping[str, Any]) -> None:
        filepath = file_info["filepath"]
        self.logger.info(f"read the file: {filepath}, size: {os.stat(filepath).st_size / (1024 ** 2)}Mb")
        with self._open_and_handle_decompress(filepath, self._get_readmode(file_info)) as f:
            if "test_stream_records" in file_info["fails"]:
                with pytest.raises(Exception) as e_info:
                    [print(r) for r in file_info["AbstractFileParser"].stream_records(f)]
                    self.logger.debug(str(e_info))
            else:
                records = [r for r in file_info["AbstractFileParser"].stream_records(f)]

                assert len(records) == file_info["num_records"]
                for index, expected_record in file_info["line_checks"].items():
                    assert records[index - 1] == expected_record



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Union

import requests
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_status import ResponseStatus


class ErrorHandler(ABC):
    @property
    @abstractmethod
    def max_retries(self) -> Union[int, None]:
        """
        Specifies maximum amount of retries for backoff policy. Return None for no limit.
        """
        pass

    @abstractmethod
    def should_retry(self, response: requests.Response) -> ResponseStatus:
        """
        Evaluate response status describing whether a failing request should be retried or ignored.
        :param response: response to evaluate
        :return: response status
        """
        pass



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import requests
from airbyte_cdk.sources.declarative.requesters.paginators.offset_paginator import OffsetPaginator
from airbyte_cdk.sources.declarative.states.dict_state import DictState

response = requests.Response()

tag = "cursor"
last_responses = [{"id": 0}, {"id": 1}]
state = DictState()


def test_return_none_if_fewer_records_than_limit():
    limit = 5
    paginator = OffsetPaginator(limit, state, tag)

    assert paginator._get_offset() == 0

    next_page_token = paginator.next_page_token(response, last_responses)

    assert next_page_token is None


def test_return_next_offset_limit_1():
    limit = 1
    paginator = OffsetPaginator(limit, state, tag)

    next_page_token = paginator.next_page_token(response, last_responses)

    assert next_page_token == {tag: 1}
    assert paginator._get_offset() == 1


def test_return_next_offset_limit_2():
    limit = 2
    paginator = OffsetPaginator(limit, state, tag)

    next_page_token = paginator.next_page_token(response, last_responses)

    assert next_page_token == {tag: 2}
    assert paginator._get_offset() == 2

    next_page_token = paginator.next_page_token(response, [{"id": 2}])
    assert next_page_token is None



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, List, Mapping, Tuple

from airbyte_cdk.models.airbyte_protocol import SyncMode
from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.source import Source


class CheckStream(ConnectionChecker):
    """
    Checks the connections by trying to read records from one or many of the streams selected by the developer
    """

    def __init__(self, stream_names: List[str]):
        self._stream_names = set(stream_names)

    def check_connection(self, source: Source, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, any]:
        streams = source.streams(config)
        stream_name_to_stream = {s.name: s for s in streams}
        if len(streams) == 0:
            return False, f"No streams to connect to from source {source}"
        for stream_name in self._stream_names:
            if stream_name in stream_name_to_stream.keys():
                stream = stream_name_to_stream[stream_name]
                try:
                    records = stream.read_records(sync_mode=SyncMode.full_refresh)
                    next(records)
                except Exception as error:
                    return False, f"Unable to connect to stream {stream} - {error}"
            else:
                raise ValueError(f"{stream_name} is not part of the catalog. Expected one of {stream_name_to_stream.keys()}")
        return True, None



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Optional, Set, Union

import requests
from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_action import ResponseAction
from airbyte_cdk.sources.streams.http.http import HttpStream


class HttpResponseFilter:
    TOO_MANY_REQUESTS_ERRORS = {429}
    DEFAULT_RETRIABLE_ERRORS = set([x for x in range(500, 600)]).union(TOO_MANY_REQUESTS_ERRORS)

    def __init__(
        self, action: Union[ResponseAction, str], *, http_codes: Set[int] = None, error_message_contain: str = None, predicate: str = ""
    ):
        """

        :param action: action to execute if a request matches
        :param http_codes: http code of matching requests
        :param error_message_contain: error substring of matching requests
        :param predicate: predicate to apply to determine if a request is matching
        """
        if isinstance(action, str):
            action = ResponseAction[action]
        self._http_codes = http_codes or set()
        self._predicate = InterpolatedBoolean(predicate)
        self._error_message_contains = error_message_contain
        self._action = action

    @property
    def action(self):
        return self._action

    def matches(self, response: requests.Response) -> Optional[ResponseAction]:
        if (
            response.status_code in self._http_codes
            or (self._response_matches_predicate(response))
            or (self._response_contains_error_message(response))
        ):
            return self._action
        else:
            return None

    def _response_matches_predicate(self, response: requests.Response) -> bool:
        return self._predicate and self._predicate.eval(None, decoded_response=response.json())

    def _response_contains_error_message(self, response: requests.Response) -> bool:
        if not self._error_message_contains:
            return False
        else:
            return self._error_message_contains in HttpStream.parse_response_error_message(response)



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass
from datetime import datetime
from functools import total_ordering


@total_ordering
@dataclass
class FileInfo:
    """Class for sharing of metadata"""

    key: str
    size: int
    last_modified: datetime

    @property
    def size_in_megabytes(self) -> float:
        """technically size in 'mebibytes'"""
        return self.size / 1024**2

    def __str__(self) -> str:
        return "Key: %s, LastModified: %s, Size: %.4fMb" % (self.key, self.last_modified.isoformat(), self.size_in_megabytes)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, FileInfo):
            return self.key == other.key and self.last_modified == other.last_modified and self.size == other.size
        return False

    def __lt__(self, other: object) -> bool:
        # comparison of less than / greater than for FileInfos is based on last modified
        if isinstance(other, FileInfo):
            return self.last_modified < other.last_modified
        else:
            raise TypeError(f"Can't compare {self.__class__} with {other.__class__}")

    def __hash__(self) -> int:
        return self.key.__hash__()



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Any, Mapping, MutableMapping, Optional, Union


class RequestOptionsProvider(ABC):
    @abstractmethod
    def request_params(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> MutableMapping[str, Any]:
        pass

    @abstractmethod
    def request_body_data(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Optional[Union[Mapping, str]]:
        pass

    @abstractmethod
    def request_body_json(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Optional[Mapping]:
        pass

    @abstractmethod
    def request_kwargs(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        pass

    @abstractmethod
    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        pass



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.states.dict_state import DictState

config = {"name": "date"}
name = "{{ config['name'] }}"
value = "{{ last_record['updated_at'] }}"
dict_mapping = {
    name: value,
}


def test_empty_state_is_none():
    state = DictState(dict_mapping, config)
    initial_state = state.get_stream_state()
    expected_state = {}
    assert expected_state == initial_state


def test_update_initial_state():
    state = DictState(dict_mapping, config)
    stream_slice = None
    stream_state = None
    last_response = {"data": {"id": "1234", "updated_at": "2021-01-01"}, "last_refresh": "2020-01-01"}
    last_record = {"id": "1234", "updated_at": "2021-01-01"}
    state.update_state(stream_slice=stream_slice, stream_state=stream_state, last_response=last_response, last_record=last_record)
    actual_state = state.get_stream_state()
    expected_state = {"date": "2021-01-01"}
    assert expected_state == actual_state


def test_update_state_with_recent_cursor():
    state = DictState(dict_mapping, config)
    stream_slice = None
    stream_state = {"date": "2020-12-31"}
    last_response = {"data": {"id": "1234", "updated_at": "2021-01-01"}, "last_refresh": "2020-01-01"}
    last_record = {"id": "1234", "updated_at": "2021-01-01"}
    state.update_state(stream_slice=stream_slice, stream_state=stream_state, last_response=last_response, last_record=last_record)
    actual_state = state.get_stream_state()
    expected_state = {"date": "2021-01-01"}
    assert expected_state == actual_state


def test_update_state_with_old_cursor():
    state = DictState(dict_mapping, config)
    stream_slice = None
    stream_state = {"date": "2021-01-02"}
    last_response = {"data": {"id": "1234", "updated_at": "2021-01-01"}, "last_refresh": "2020-01-01"}
    last_record = {"id": "1234", "updated_at": "2021-01-01"}
    state.update_state(stream_slice=stream_slice, stream_state=stream_state, last_response=last_response, last_record=last_record)
    actual_state = state.get_stream_state()
    expected_state = {"date": "2021-01-02"}
    assert expected_state == actual_state


def test_update_state_with_older_state():
    state = DictState(dict_mapping, config)
    stream_slice = None
    stream_state = {"date": "2021-01-02"}
    last_response = {"data": {"id": "1234", "updated_at": "2021-01-02"}, "last_refresh": "2020-01-01"}
    last_record = {"id": "1234", "updated_at": "2021-01-02"}
    state.update_state(stream_slice=stream_slice, stream_state=stream_state, last_response=last_response, last_record=last_record)
    actual_state = state.get_stream_state()
    expected_state = {"date": "2021-01-02"}

    out_of_order_response = {"data": {"id": "1234", "updated_at": "2021-01-02"}, "last_refresh": "2020-01-01"}
    out_of_order_record = {"id": "1234", "updated_at": "2021-01-01"}
    state.update_state(
        stream_slice=stream_slice, stream_state=stream_state, last_response=out_of_order_response, last_record=out_of_order_record
    )
    assert expected_state == actual_state


def test_state_is_a_timestamp():
    state = DictState(dict_mapping, config)
    stream_slice = None
    stream_state = {"date": 12345}
    last_response = {"data": {"id": "1234", "updated_at": 123456}, "last_refresh": "2020-01-01"}
    last_record = {"id": "1234", "updated_at": 123456}
    state.update_state(stream_slice=stream_slice, stream_state=stream_state, last_response=last_response, last_record=last_record)
    actual_state = state.get_stream_state()
    expected_state = {"date": 123456}
    assert expected_state == actual_state



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from .file_info import FileInfo
from .files_spec import FilesSpec
from .files_stream import FilesStream, IncrementalFilesStream
from .storage_file import StorageFile

__all__ = ["FileInfo", "FilesStream", "IncrementalFilesStream", "FilesSpec", "StorageFile"]



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.sources.declarative.parsers.undefined_reference_exception import UndefinedReferenceException
from airbyte_cdk.sources.declarative.parsers.yaml_parser import YamlParser

parser = YamlParser()


def test():
    content = """
    limit: 50
    """
    config = parser.parse(content)
    assert config["limit"] == 50


def test_get_ref():
    s = """
    limit_ref: "*ref(limit)"
    """
    ref_key = parser.get_ref_key(s)
    assert ref_key == "limit"


def test_get_ref_no_ref():
    s = """
    limit: 50
    """
    ref_key = parser.get_ref_key(s)
    assert ref_key is None


def test_refer():
    content = """
    limit: 50
    limit_ref: "*ref(limit)"
    """
    config = parser.parse(content)
    assert config["limit_ref"] == 50


def test_refer_to_inner():
    content = """
    dict:
      limit: 50
    limit_ref: "*ref(dict.limit)"
    """
    config = parser.parse(content)
    assert config["limit_ref"] == 50


def test_refer_to_non_existant_struct():
    content = """
    dict:
      limit: 50
    limit_ref: "*ref(not_dict)"
    """
    with pytest.raises(UndefinedReferenceException):
        parser.parse(content)


def test_refer_in_dict():
    content = """
    limit: 50
    offset_request_parameters:
      offset: "{{ next_page_token['offset'] }}"
      limit: "*ref(limit)"
    """
    config = parser.parse(content)
    assert config["offset_request_parameters"]["offset"] == "{{ next_page_token['offset'] }}"
    assert config["offset_request_parameters"]["limit"] == 50


def test_refer_to_dict():
    content = """
    limit: 50
    offset_request_parameters:
      offset: "{{ next_page_token['offset'] }}"
      limit: "*ref(limit)"
    offset_pagination_request_parameters:
      class: InterpolatedRequestParameterProvider
      request_parameters: "*ref(offset_request_parameters)"
    """
    config = parser.parse(content)
    assert config["limit"] == 50
    assert config["offset_request_parameters"]["limit"] == 50
    assert len(config["offset_pagination_request_parameters"]) == 2
    assert config["offset_pagination_request_parameters"]["request_parameters"]["limit"] == 50
    assert config["offset_pagination_request_parameters"]["request_parameters"]["offset"] == "{{ next_page_token['offset'] }}"


def test_refer_and_overwrite():
    content = """
    limit: 50
    custom_limit: 25
    offset_request_parameters:
      offset: "{{ next_page_token['offset'] }}"
      limit: "*ref(limit)"
    custom_request_parameters:
      ref: "*ref(offset_request_parameters)"
      limit: "*ref(custom_limit)"
    """
    config = parser.parse(content)
    assert config["offset_request_parameters"]["limit"] == 50
    assert config["custom_request_parameters"]["limit"] == 25

    assert config["offset_request_parameters"]["offset"] == "{{ next_page_token['offset'] }}"
    assert config["custom_request_parameters"]["offset"] == "{{ next_page_token['offset'] }}"


def test_collision():
    content = """
example:
  nested:
    path: "first one"
    more_nested:
        value: "found it!"
  nested.path: "uh oh"
reference_to_nested_path:
  ref: "*ref(example.nested.path)"
reference_to_nested_nested_value:
  ref: "*ref(example.nested.more_nested.value)"
    """
    config = parser.parse(content)
    assert config["example"]["nested"]["path"] == "first one"
    assert config["example"]["nested.path"] == "uh oh"
    assert config["reference_to_nested_path"] == "uh oh"
    assert config["example"]["nested"]["more_nested"]["value"] == "found it!"
    assert config["reference_to_nested_nested_value"] == "found it!"


def test_list():
    content = """
    list:
      - "A"
      - "B"
    elem_ref: "*ref(list[0])"
    """
    config = parser.parse(content)
    elem_ref = config["elem_ref"]
    assert elem_ref == "A"



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


import logging
import time
from abc import ABC, abstractmethod
from typing import Any, Iterator, List, Mapping
from uuid import uuid4

import pytest
from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.streams.files import FilesStream
from airbyte_cdk.sources.streams.files.formats.csv_parser import CsvParser
from .sample_files import SAMPLE_DIR


LOGGER = logging.getLogger("airbyte")
JSONTYPE_TO_PYTHONTYPE = {"string": str, "number": float, "integer": int, "object": dict, "array": list, "boolean": bool, "null": None}


class AbstractFilesStreamIntegrationTest(ABC):
    """Prefix this class with Abstract so the tests don't run here but only in the children"""

    temp_bucket_prefix = "airbytetest-"

    @pytest.fixture(scope="session")
    def cloud_bucket_prefix(self) -> str:
        return self.temp_bucket_prefix

    @pytest.fixture(scope="session")
    def format(self) -> Mapping[str, Any]:
        return {"filetype": "csv"}

    @pytest.fixture(scope="session")
    def airbyte_system_columns(self) -> Mapping[str, str]:
        return {FilesStream.ab_additional_col: "object", FilesStream.ab_last_mod_col: "string", FilesStream.ab_file_name_col: "string"}

    @property
    @abstractmethod
    def stream_class(self) -> type:
        """
        :return: provider specific FilesStream class
        """

    @property
    @abstractmethod
    def credentials(self) -> Mapping:
        """
        These will be added automatically to the provider property

        :return: mapping of provider specific credentials
        """

    @abstractmethod
    def provider(self, bucket_name: str) -> Mapping:
        """
        :return: provider specific provider dict as described in spec.json
                 leave out the credentials, they will be added from the credentials() method
        """

    @abstractmethod
    def cloud_files(self, cloud_bucket_name: str, credentials: Mapping, files_to_upload: List, private: bool = True) -> Iterator[str]:
        """
        See source-s3 for example what the override of this needs to achieve.

        :param cloud_bucket_name: name of bucket (or equivalent)
        :param credentials: mapping of provider specific credentials
        :param files_to_upload: list of paths to local files to upload, pass empty list to test zero files case
        :param private: whether or not to make the files private and require credentials to read, defaults to True
        :yield: url filepath to uploaded file
        """

    @abstractmethod
    def teardown_infra(self, cloud_bucket_name: str, credentials: Mapping) -> None:
        """
        Provider-specific logic to tidy up any cloud resources.
        See source-s3 for example.

        :param cloud_bucket_name: bucket (or equivalent) name
        :param credentials: mapping of provider specific credentials
        """

    def _stream_records_test_logic(
        self,
        cloud_bucket_name: str,
        format: Mapping[str, str],
        airbyte_system_columns: Mapping[str, str],
        sync_mode: Any,
        files: List[str],
        path_pattern: str,
        private: bool,
        num_columns: Any,
        num_records: Any,
        expected_schema: Mapping[str, Any],
        user_schema: Mapping[str, Any],
        fails: Any,
        state: Any = None,
    ) -> Any:
        uploaded_files = [fpath for fpath in self.cloud_files(cloud_bucket_name, self.credentials, files, private)]
        LOGGER.info(f"file(s) uploaded: {uploaded_files}")

        # emulate state for incremental testing
        # since we're not actually saving state out to file here, we pass schema in to our FileStream creation...
        # this isn't how it will work in Airbyte but it's a close enough emulation
        current_state = state if state is not None else {FilesStream.ab_last_mod_col: "1970-01-01T00:00:00+0000"}
        if (user_schema is None) and ("schema" in current_state.keys()):
            user_schema = current_state["schema"]

        full_expected_schema = {**expected_schema, **airbyte_system_columns}
        str_user_schema = str(user_schema).replace("'", '"') if user_schema is not None else None
        total_num_columns = num_columns + len(airbyte_system_columns.keys())
        provider = {**self.provider(cloud_bucket_name), **self.credentials} if private else self.provider(cloud_bucket_name)

        if not fails:
            stream = self.stream_class("dataset", provider, format, path_pattern, str_user_schema)
            LOGGER.info(f"Testing stream_records() in SyncMode:{sync_mode.value}")

            # check we return correct schema from get_json_schema()
            assert stream._get_schema_map() == full_expected_schema

            records = []
            for stream_slice in stream.stream_slices(sync_mode=sync_mode, stream_state=current_state):
                if stream_slice is not None:
                    # we need to do this in order to work out which extra columns (if any) we expect in this stream_slice
                    expected_columns = []
                    for file_dict in stream_slice["files"]:
                        file_reader = CsvParser(format)
                        with file_dict["storage_file"].open(file_reader.is_binary) as f:
                            expected_columns.extend(list(file_reader.get_inferred_schema(f).keys()))
                    expected_columns = set(expected_columns)  # de-dupe

                    for record in stream.read_records(sync_mode, stream_slice=stream_slice):
                        # check actual record values match expected schema
                        assert all(
                            [
                                isinstance(record[col], JSONTYPE_TO_PYTHONTYPE[typ]) or record[col] is None
                                for col, typ in full_expected_schema.items()
                            ]
                        )
                        records.append(record)

            assert all([len(r.keys()) == total_num_columns for r in records])
            assert len(records) == num_records

            # check additional properties included as expected if any exist
            if (user_schema is not None) and (expected_columns != set(user_schema.keys())):
                for additional_property in expected_columns.difference(set(user_schema.keys())):
                    # since we can't be dynamically aware of which records should have which additional props, we just any() check here
                    assert any([additional_property in r[FilesStream.ab_additional_col].keys() for r in records])

            # returning state by simulating call to get_updated_state() with final record so we can test incremental
            return stream.get_updated_state(current_stream_state=current_state, latest_record=records[-1])

        else:
            with pytest.raises(Exception) as e_info:
                stream = self.stream_class("dataset", provider, format, path_pattern, str_user_schema)
                LOGGER.info(f"Testing EXPECTED FAILURE stream_records() in SyncMode:{sync_mode.value}")

                stream.get_json_schema()

                records = []
                for stream_slice in stream.stream_slices(sync_mode=sync_mode, stream_state=current_state):
                    for record in stream.read_records(sync_mode, stream_slice=stream_slice):
                        records.append(record)

                LOGGER.info(f"Failed as expected, error: {e_info}")

    @pytest.mark.parametrize(
        # make user_schema None to test auto-inference. Exclude any _airbyte system columns in expected_schema.
        "files, path_pattern, private, num_columns, num_records, expected_schema, user_schema, incremental, fails",
        [
            # single file tests
            (  # public
                [SAMPLE_DIR.joinpath("simple_test.csv")],
                "**",
                False,
                3,
                8,
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                False,
                False,
            ),
            (  # private
                [SAMPLE_DIR.joinpath("simple_test.csv")],
                "**",
                True,
                3,
                8,
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                False,
                False,
            ),
            (  # provided schema exact match to actual schema
                [SAMPLE_DIR.joinpath("simple_test.csv")],
                "**",
                True,
                3,
                8,
                {"id": "integer", "name": "string", "valid": "boolean"},
                {"id": "integer", "name": "string", "valid": "boolean"},
                False,
                False,
            ),
            (  # provided schema not matching datatypes, expect successful coercion
                [SAMPLE_DIR.joinpath("simple_test.csv")],
                "**",
                True,
                3,
                8,
                {"id": "string", "name": "string", "valid": "string"},
                {"id": "string", "name": "string", "valid": "string"},
                False,
                False,
            ),
            (  # provided incompatible schema, expect fail
                [SAMPLE_DIR.joinpath("simple_test.csv")],
                "**",
                True,
                3,
                8,
                {"id": "boolean", "name": "boolean", "valid": "boolean"},
                {"id": "boolean", "name": "boolean", "valid": "boolean"},
                False,
                True,
            ),
            # multiple file tests (all have identical schemas)
            (  # public, auto-infer
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                False,
                3,
                17,
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                False,
                False,
            ),
            (  # private, auto-infer
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                True,
                3,
                17,
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                False,
                False,
            ),
            (  # provided schema exact match to actual schema
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                True,
                3,
                17,
                {"id": "integer", "name": "string", "valid": "boolean"},
                {"id": "integer", "name": "string", "valid": "boolean"},
                False,
                False,
            ),
            (  # provided schema not matching datatypes, expect successful coercion
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                True,
                3,
                17,
                {"id": "string", "name": "string", "valid": "string"},
                {"id": "string", "name": "string", "valid": "string"},
                False,
                False,
            ),
            (  # provided incompatible schema, expect fail
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                True,
                3,
                17,
                {"id": "boolean", "name": "boolean", "valid": "boolean"},
                {"id": "boolean", "name": "boolean", "valid": "boolean"},
                False,
                True,
            ),
            # multiple file tests (different but merge-able schemas)
            (  # auto-infer
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_1.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_2.csv"),
                ],
                "**",
                True,
                6,
                17,
                {"id": "integer", "name": "string", "valid": "boolean", "location": "string", "percentage": "number", "nullable": "string"},
                None,
                False,
                False,
            ),
            (  # provided schema, not containing all columns (extra columns should go into FileStream.ab_additional_col)
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_1.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_2.csv"),
                ],
                "**",
                True,
                3,
                17,
                {"id": "integer", "name": "string", "valid": "boolean"},
                {"id": "integer", "name": "string", "valid": "boolean"},
                False,
                False,
            ),
            # pattern matching tests with additional files present that we don't want to read
            (  # at top-level of bucket
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("file_to_skip.csv"),
                    SAMPLE_DIR.joinpath("file_to_skip.txt"),
                ],
                "simple*",
                True,
                3,
                11,
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                False,
                False,
            ),
            (  # at multiple levels of bucket
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("file_to_skip.csv"),
                    SAMPLE_DIR.joinpath("file_to_skip.txt"),
                    SAMPLE_DIR.joinpath("pattern_match_test/this_folder/simple_test.csv"),
                    SAMPLE_DIR.joinpath("pattern_match_test/not_this_folder/file_to_skip.csv"),
                    SAMPLE_DIR.joinpath("pattern_match_test/not_this_folder/file_to_skip.txt"),
                ],
                "**/simple*",
                True,
                3,
                19,
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                False,
                False,
            ),
            # incremental tests (passing num_records/num_columns/fails as lists holding value for each file in order)
            (  # auto-infer, all same schema
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                True,
                [3, 3, 3],
                [8, 3, 6],
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                True,
                [False, False, False],
            ),
            (  # provided schema, all same schema
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("simple_test_2.csv"),
                    SAMPLE_DIR.joinpath("simple_test_3.csv"),
                ],
                "**",
                True,
                [3, 3, 3],
                [8, 3, 6],
                {"id": "integer", "name": "string", "valid": "boolean"},
                {"id": "integer", "name": "string", "valid": "boolean"},
                True,
                [False, False, False],
            ),
            (  # auto-infer, (different but merge-able schemas)
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_1.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_2.csv"),
                ],
                "**",
                True,
                [3, 3, 3],
                [8, 3, 6],
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                True,
                [False, False, False],
            ),
            (  # same as previous but change order and expect 5 columns instead of 3 in all
                [
                    SAMPLE_DIR.joinpath("multi_file_diffschema_2.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_1.csv"),
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                ],
                "**",
                True,
                [5, 5, 5],
                [6, 3, 8],
                {"id": "integer", "name": "string", "valid": "boolean", "percentage": "number", "nullable": "string"},
                None,
                True,
                [False, False, False],
            ),
            (  # like previous test but with a user_schema limiting columns
                [
                    SAMPLE_DIR.joinpath("multi_file_diffschema_2.csv"),
                    SAMPLE_DIR.joinpath("multi_file_diffschema_1.csv"),
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                ],
                "**",
                True,
                [2, 2, 2],
                [6, 3, 8],
                {"id": "integer", "name": "string"},
                {"id": "integer", "name": "string"},
                True,
                [False, False, False],
            ),
            (  # fail when 2nd file has incompatible schema, auto-infer
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("incompatible_schema.csv"),
                ],
                "**",
                True,
                [3, 3],
                [8, 8],
                {"id": "integer", "name": "string", "valid": "boolean"},
                None,
                True,
                [False, True],
            ),
            (  # fail when 2nd file has incompatible schema, provided schema
                [
                    SAMPLE_DIR.joinpath("simple_test.csv"),
                    SAMPLE_DIR.joinpath("incompatible_schema.csv"),
                ],
                "**",
                True,
                [3, 3],
                [8, 8],
                {"id": "integer", "name": "string", "valid": "boolean"},
                {"id": "integer", "name": "string", "valid": "boolean"},
                True,
                [False, True],
            ),
        ],
    )
    def test_stream_records(
        self,
        cloud_bucket_prefix: str,
        format: Mapping[str, Any],
        airbyte_system_columns: Mapping[str, str],
        files: List[str],
        path_pattern: str,
        private: bool,
        num_columns: List[int],
        num_records: List[int],
        expected_schema: Mapping[str, Any],
        user_schema: Mapping[str, Any],
        incremental: bool,
        fails: List[bool],
    ) -> None:
        try:
            if not incremental:  # we expect matching behaviour here in either sync_mode
                for sync_mode in [
                    SyncMode("full_refresh"),
                    SyncMode("incremental"),
                ]:
                    cloud_bucket_name = f"{cloud_bucket_prefix}{uuid4()}"
                    self._stream_records_test_logic(
                        cloud_bucket_name,
                        format,
                        airbyte_system_columns,
                        sync_mode,
                        files,
                        path_pattern,
                        private,
                        num_columns,
                        num_records,
                        expected_schema,
                        user_schema,
                        fails,
                    )
                    self.teardown_infra(cloud_bucket_name, self.credentials)
            else:
                cloud_bucket_name = f"{cloud_bucket_prefix}{uuid4()}"
                latest_state = None
                for i in range(len(files)):
                    latest_state = self._stream_records_test_logic(
                        cloud_bucket_name,
                        format,
                        airbyte_system_columns,
                        SyncMode("incremental"),
                        [files[i]],
                        path_pattern,
                        private,
                        num_columns[i],
                        num_records[i],
                        expected_schema,
                        user_schema,
                        fails[i],
                        state=latest_state,
                    )
                    LOGGER.info(f"incremental state: {latest_state}")
                    # small delay to ensure next file gets later last_modified timestamp
                    time.sleep(1)
                self.teardown_infra(cloud_bucket_name, self.credentials)

        except Exception as e:
            self.teardown_infra(cloud_bucket_name, self.credentials)
            raise e



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import List, Union

import airbyte_cdk.sources.declarative.requesters.error_handlers.response_status as response_status
import requests
from airbyte_cdk.sources.declarative.requesters.error_handlers.error_handler import ErrorHandler
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_action import ResponseAction
from airbyte_cdk.sources.declarative.requesters.error_handlers.response_status import ResponseStatus


class CompositeErrorHandler(ErrorHandler):
    """
    Sample config chaining 2 different retriers:
        error_handler:
          type: "CompositeErrorHandler"
          error_handlers:
            - response_filters:
                - predicate: "{{ 'codase' in decoded_response }}"
                  action: RETRY
              backoff_strategies:
                - type: "ConstantBackoffStrategy"
                  backoff_time_in_seconds: 5
            - response_filters:
                - http_codes: [ 403 ]
                  action: RETRY
              backoff_strategies:
                - type: "ConstantBackoffStrategy"
                  backoff_time_in_seconds: 10
    """

    def __init__(self, error_handlers: List[ErrorHandler]):
        """

        :param error_handlers: list of error handlers
        """
        self._error_handlers = error_handlers
        if not self._error_handlers:
            raise ValueError("CompositeErrorHandler expects at least 1 underlying error handler")

    @property
    def max_retries(self) -> Union[int, None]:
        return self._error_handlers[0].max_retries

    def should_retry(self, response: requests.Response) -> ResponseStatus:
        should_retry = None
        for retrier in self._error_handlers:
            should_retry = retrier.should_retry(response)
            if should_retry.action == ResponseAction.SUCCESS:
                return response_status.SUCCESS
            if should_retry == response_status.IGNORE or should_retry.action == ResponseAction.RETRY:
                return should_retry
        return should_retry



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, List, Mapping, Optional

import requests
from airbyte_cdk.sources.declarative.requesters.paginators.paginator import Paginator
from airbyte_cdk.sources.declarative.states.dict_state import DictState


class OffsetPaginator(Paginator):
    def __init__(self, page_size: int, state: Optional[DictState] = None, offset_key: str = "offset"):
        self._limit = page_size
        self._state = state or DictState()
        self._offsetKey = offset_key
        self._update_state_with_offset(0)

    def next_page_token(self, response: requests.Response, last_records: List[Mapping[str, Any]]) -> Optional[Mapping[str, Any]]:
        if not last_records or len(last_records) < self._limit:
            return None
        offset = self._get_offset() + self._limit
        token_map = {self._offsetKey: offset}
        self._update_state_with_offset(offset)
        return token_map

    def _update_state_with_offset(self, offset):
        self._state.update_state(**{self._offsetKey: offset})

    def _get_offset(self):
        return self._state.get_state(self._offsetKey)



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Any, List, Mapping

import requests
from airbyte_cdk.sources.declarative.types import Record


class HttpSelector(ABC):
    @abstractmethod
    def select_records(
        self,
        response: requests.Response,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> List[Record]:
        pass



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


import pathlib

from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

main_deps = [
    "backoff",
    "dpath~=2.0.1",
    "jsonschema~=3.2.0",
    "jsonref~=0.2",
    "pendulum",
    "pydantic~=1.6",
    "PyYAML~=5.4",
    "requests",
    "vcrpy",
    "Deprecated~=1.2",
    "Jinja2~=3.1.2",
    "jello~=1.5.2",
]
files_source_deps = [
    "pyarrow~=8.0.0",
    "wcmatch~=8.2",
    "dill~=0.3.5",
    "fastavro~=1.4.11",
    "python-snappy~=0.6.1",
    "pandas~=1.3.1",
    "avro~=1.11.0",
]
sphinx_deps = ["Sphinx~=4.2", "sphinx-rtd-theme~=1.0"]
test_deps = ["MyPy~=0.812", "pytest", "pytest-cov", "pytest-mock", "pytest-order", "requests-mock", "pytest-httpserver"]

dev_deps = main_deps + files_source_deps + sphinx_deps + test_deps

setup(
    name="airbyte-cdk-PHLAIR",
    version="0.2.2",
    description="A framework for writing Airbyte Connectors.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Airbyte",
    author_email="contact@airbyte.io",
    license="MIT",
    url="https://github.com/airbytehq/airbyte",
    classifiers=[
        # This information is used when browsing on PyPi.
        # Dev Status
        "Development Status :: 3 - Alpha",
        # Project Audience
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        # Python Version Support
        "Programming Language :: Python :: 3.9",
    ],
    keywords="airbyte connector-development-kit cdk",
    project_urls={
        "Documentation": "https://docs.airbyte.io/",
        "Source": "https://github.com/airbytehq/airbyte",
        "Tracker": "https://github.com/airbytehq/airbyte/issues",
    },
    packages=find_packages(exclude=("unit_tests",)),
    install_requires=main_deps,
    python_requires=">=3.9",
    extras_require={"files-source": files_source_deps, "sphinx-docs": sphinx_deps, "dev": dev_deps},
)



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Optional

from pydantic import BaseModel, Field


class CsvFormat(BaseModel):
    'This connector utilises <a href="https: // arrow.apache.org/docs/python/generated/pyarrow.csv.open_csv.html" target="_blank">PyArrow (Apache Arrow)</a> for CSV parsing.'

    class Config:
        title = "CSV"

    filetype: str = Field(
        "csv",
        const=True,
    )

    delimiter: str = Field(
        default=",",
        min_length=1,
        description="The character delimiting individual cells in the CSV data. This may only be a 1-character string. For tab-delimited data enter '\\t'.",
        order=0,
    )
    infer_datatypes: Optional[bool] = Field(
        default=True,
        description="Configures whether a schema for the source should be inferred from the current data or not. "
        "If set to false and a custom schema is set, then the manually enforced schema is used. "
        "If a schema is not manually set, and this is set to false, then all fields will be read as strings",
        order=1,
    )
    quote_char: str = Field(
        title="Quote Character",
        default='"',
        description="The character used for quoting CSV values. To disallow quoting, make this field blank.",
        order=2,
    )
    escape_char: Optional[str] = Field(
        title="Escape Character",
        default=None,
        description="The character used for escaping special characters. To disallow escaping, leave this field blank.",
        order=3,
    )
    encoding: Optional[str] = Field(
        default="utf8",
        description='The character encoding of the CSV data. Leave blank to default to <strong>UTF8</strong>. See <a href="https://docs.python.org/3/library/codecs.html#standard-encodings" target="_blank">list of python encodings</a> for allowable options.',
        order=4,
    )
    double_quote: bool = Field(
        default=True, description="Whether two quotes in a quoted CSV value denote a single quote in the data.", order=5
    )
    newlines_in_values: bool = Field(
        title="Allow newlines in values",
        default=False,
        description="Whether newline characters are allowed in CSV values. Turning this on may affect performance. Leave blank to default to False.",
        order=6,
    )
    additional_reader_options: str = Field(
        default="{}",
        description='Optionally add a valid JSON string here to provide additional options to the csv reader. Mappings must correspond to options <a href="https://arrow.apache.org/docs/python/generated/pyarrow.csv.ConvertOptions.html#pyarrow.csv.ConvertOptions" target="_blank">detailed here</a>. \'column_types\' is used internally to handle schema so overriding that would likely cause problems.',
        examples=[
            '{"timestamp_parsers": ["%m/%d/%Y %H:%M", "%Y/%m/%d %H:%M"], "strings_can_be_null": true, "null_values": ["NA", "NULL"]}'
        ],
        order=7,
    )
    advanced_options: str = Field(
        default="{}",
        description="Optionally add a valid JSON string here to provide additional <a href=\"https://arrow.apache.org/docs/python/generated/pyarrow.csv.ReadOptions.html#pyarrow.csv.ReadOptions\" target=\"_blank\">Pyarrow ReadOptions</a>. Specify 'column_names' here if your CSV doesn't have header, or if you want to use custom column names. 'block_size' and 'encoding' are already used above, specify them again here will override the values above.",
        examples=['{"column_names": ["column1", "column2"]}'],
        order=8,
    )
    block_size: int = Field(
        default=10000,
        description="The chunk size in bytes to process at a time in memory from each file. If your data is particularly wide and failing during schema detection, increasing this should solve it. Beware of raising this too high as you could hit OOM errors.",
        order=9,
    )



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import json

import pytest
import requests
from airbyte_cdk.sources.declarative.decoders.json_decoder import JsonDecoder
from airbyte_cdk.sources.declarative.requesters.paginators.conditional_paginator import ConditionalPaginator
from airbyte_cdk.sources.declarative.states.dict_state import DictState


@pytest.mark.parametrize(
    "test_name, stop_condition_template, expected_next_page_token, expected_page",
    [
        ("test_stop_pagination_from_response_body", "{{ not decoded_response['accounts'] }}", None, 1),
        ("test_stop_pagination_from_config", "{{ config['response_override'] == decoded_response['_metadata']['content'] }}", None, 1),
        ("test_continue_pagination_from_response_body", "{{ decoded_response['end'] == decoded_response['total'] - 1 }}", {"page": 2}, 2),
        ("test_continue_pagination_from_response_headers", "{{ decoded_response['headers']['has_more'] }}", {"page": 2}, 2),
        ("test_continue_pagination_from_last_records", "{{ last_records[-1]['more_records'] == False }}", {"page": 2}, 2),
        ("test_continue_pagination_for_empty_dict_evaluates_false", "{{ decoded_response['characters'] }}", {"page": 2}, 2),
    ],
)
def test_interpolated_request_header(test_name, stop_condition_template, expected_next_page_token, expected_page):
    state = DictState()
    state.update_state(page=1)
    decoder = JsonDecoder()
    config = {"response_override": "stop_if_you_see_me"}

    response = requests.Response()
    response.headers = {"has_more": True}
    response_body = {"_metadata": {"content": "stop_if_you_see_me"}, "accounts": [], "end": 99, "total": 200, "characters": {}}
    response._content = json.dumps(response_body).encode("utf-8")
    last_records = [{"id": 0, "more_records": True}, {"id": 1, "more_records": True}]

    paginator = ConditionalPaginator(stop_condition_template, state, decoder, config)
    next_page_token = paginator.next_page_token(response, last_records)

    assert next_page_token == expected_next_page_token
    assert state.get_state("page") == expected_page



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass
from typing import Any, List, Mapping, Union

import dpath.util
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.transformations import RecordTransformation
from airbyte_cdk.sources.declarative.types import Config, FieldPointer, StreamSlice, StreamState


@dataclass(frozen=True)
class AddedFieldDefinition:
    path: FieldPointer
    value: Union[InterpolatedString, str]


@dataclass(frozen=True)
class ParsedAddFieldDefinition:
    path: FieldPointer
    value: InterpolatedString


class AddFields(RecordTransformation):
    """
    Transformation which adds field to an output record. The path of the added field can be nested. Adding nested fields will create all
    necessary parent objects (like mkdir -p). Adding fields to an array will extend the array to that index (filling intermediate
    indices with null values). So if you add a field at index 5 to the array ["value"], it will become ["value", null, null, null, null,
    "new_value"].


    This transformation has access to the following contextual values:
        record: the record about to be output by the connector
        config: the input configuration provided to a connector
        stream_state: the current state of the stream
        stream_slice: the current stream slice being read



    Examples of instantiating this transformation via YAML:
    - type: AddFields
      fields:
        # hardcoded constant
        - path: ["path"]
          value: "static_value"

        # nested path
        - path: ["path", "to", "field"]
          value: "static"

        # from config
        - path: ["shop_id"]
          value: "{{ config.shop_id }}"

        # from state
        - path: ["current_state"]
          value: "{{ stream_state.cursor_field }}" # Or {{ stream_state['cursor_field'] }}

        # from record
        - path: ["unnested_value"]
          value: {{ record.nested.field }}

        # from stream_slice
        - path: ["start_date"]
          value: {{ stream_slice.start_date }}

        # by supplying any valid Jinja template directive or expression https://jinja.palletsprojects.com/en/3.1.x/templates/#
        - path: ["two_times_two"]
          value: {{ 2 * 2 }}
    """

    def __init__(self, fields: List[AddedFieldDefinition]):
        self._fields: List[ParsedAddFieldDefinition] = []
        for field in fields:
            if len(field.path) < 1:
                raise f"Expected a non-zero-length path for the AddFields transformation {field}"

            if not isinstance(field.value, InterpolatedString):
                if not isinstance(field.value, str):
                    raise f"Expected a string value for the AddFields transformation: {field}"
                else:
                    self._fields.append(ParsedAddFieldDefinition(field.path, InterpolatedString(field.value)))
            else:
                self._fields.append(ParsedAddFieldDefinition(field.path, field.value))

    def transform(
        self, record: Mapping[str, Any], config: Config = None, stream_state: StreamState = None, stream_slice: StreamSlice = None
    ) -> Mapping[str, Any]:
        kwargs = {"record": record, "stream_state": stream_state, "stream_slice": stream_slice}
        for field in self._fields:
            value = field.value.eval(config, **kwargs)
            dpath.util.new(record, field.path, value)

        return record

    def __eq__(self, other):
        return self.__dict__ == other.__dict__



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


import os
import random
import string
import tempfile
from typing import Iterator

SAMPLE_DIR = os.path.join(tempfile.gettempdir(), "sample_files")
SCHEMA_CORE = {"name": str}
SCHEMA_1 = {**SCHEMA_CORE,**{"valid": bool}}
SCHEMA_DIFF_1 = {**SCHEMA_CORE,**{"location": str}}
SCHEMA_DIFF_2 = {**SCHEMA_1,**{"percentage": float, "nullable": None}}
SCHEMA_INCOMPATIBLE = dict(SCHEMA_1, valid=str)
STRING_LENGTH = 10

def _generate_string(length: int):
    return ''.join(random.choice(string.ascii_lowercase) for i in range(length))

def _generate_bool():
    return random.choice([True, False])

def _generate_float():
    sign = random.choice([-1, 1])
    magnitude = random.randint(1, 100)
    return round(sign * magnitude * random.random(), 2)

def _data_generator(start_id: int, num_rows: int, schema: dict) -> Iterator:
    yield ["id"] + list(schema.keys())
    for x in range(num_rows):
        row = [start_id+x]
        for typ in schema.values():
            if typ is str:
                row.append(_generate_string(STRING_LENGTH))
            elif typ is bool:
                row.append(_generate_bool())
            elif typ is float:
                row.append(_generate_float())
            else:
                row.append(None)
        yield(row)

def _create_file(dirpath, filename):
    if not os.path.exists(dirpath):
        print(f"creating dir(s): {dirpath}")
        os.makedirs(dirpath, exist_ok=True)
    filepath = os.path.join(dirpath, filename)
    print(f"creating file: {filepath}")
    open(filepath, "w+").close()
    return filepath

def _format_row_for_csv(row: list):
    return ",".join([str(x) for x in row]) + "\n"

def _append_csv_row_to_file(csv_row: str, filepath: str):
    with open(filepath, "a") as f:
        f.write(csv_row)

def _build_data_file(name: str, start_id: int, num_rows: int, schema: dict, custom_dir: str=None):
    dir = custom_dir if custom_dir is not None else SAMPLE_DIR
    fpath = _create_file(dir, name)
    for row in _data_generator(start_id, num_rows, schema):
        _append_csv_row_to_file(_format_row_for_csv(row), fpath)

def generate_sample_files():
    _build_data_file("simple_test.csv", start_id=1, num_rows=8, schema=SCHEMA_1)
    _build_data_file("simple_test_2.csv", start_id=9, num_rows=3, schema=SCHEMA_1)
    _build_data_file("simple_test_3.csv", start_id=12, num_rows=6, schema=SCHEMA_1)
    _build_data_file("no_rows.csv", start_id=1, num_rows=0, schema=SCHEMA_1)
    _build_data_file("multi_file_diffschema_1.csv", start_id=9, num_rows=3, schema=SCHEMA_DIFF_1)
    _build_data_file("multi_file_diffschema_2.csv", start_id=12, num_rows=6, schema=SCHEMA_DIFF_2)
    _build_data_file("incompatible_schema.csv", start_id=1, num_rows=8, schema=SCHEMA_INCOMPATIBLE)
    _build_data_file("file_to_skip.csv", start_id=1, num_rows=2, schema=SCHEMA_1)
    _build_data_file("file_to_skip.txt", start_id=1, num_rows=2, schema=SCHEMA_1)
    _build_data_file("simple_test.csv", start_id=1, num_rows=8, schema=SCHEMA_1,
                    custom_dir=os.path.join(SAMPLE_DIR, "pattern_match_test/this_folder"))
    _build_data_file("file_to_skip.csv", start_id=1, num_rows=2, schema=SCHEMA_1,
                    custom_dir=os.path.join(SAMPLE_DIR, "pattern_match_test/not_this_folder"))
    _build_data_file("file_to_skip.txt", start_id=1, num_rows=2, schema=SCHEMA_1,
                    custom_dir=os.path.join(SAMPLE_DIR, "pattern_match_test/not_this_folder"))



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

from typing import Any, List, Mapping

from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.types import Config, Record


class RecordFilter:
    def __init__(self, config: Config, condition: str = None):
        self._config = config
        self._filter_interpolator = InterpolatedBoolean(condition)

    def filter_records(
        self,
        records: List[Record],
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> List[Record]:
        kwargs = {"stream_state": stream_state, "stream_slice": stream_slice, "next_page_token": next_page_token}
        return [record for record in records if self._filter_interpolator.eval(self._config, record=record, **kwargs)]



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#


import json
import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from datetime import datetime, timedelta
from functools import lru_cache
from traceback import format_exc
from typing import Any, Dict, Iterable, Iterator, List, Mapping, MutableMapping, Optional, Union

from airbyte_cdk.models.airbyte_protocol import SyncMode
from airbyte_cdk.sources.streams import Stream
from wcmatch.glob import GLOBSTAR, SPLIT, globmatch

from .file_info import FileInfo
from .formats import AbstractFileParser, AvroParser, CsvParser, ParquetParser
from .storage_file import StorageFile

JSON_TYPES = ["string", "number", "integer", "object", "array", "boolean", "null"]

FORMAT_PARSER_MAP = {
    "csv": CsvParser,
    "parquet": ParquetParser,
    "avro": AvroParser,
}

LOGGER = logging.getLogger("airbyte")


class ConfigurationError(Exception):
    """Client mis-configured"""


class FilesStream(Stream, ABC):

    # TODO: make these user configurable in spec.json
    ab_additional_col = "_ab_additional_properties"
    ab_last_mod_col = "_ab_source_file_last_modified"
    ab_file_name_col = "_ab_source_file_url"
    airbyte_columns = [ab_additional_col, ab_last_mod_col, ab_file_name_col]
    datetime_format_string = "%Y-%m-%dT%H:%M:%S%z"

    def __init__(self, dataset: str, provider: dict, format: dict, path_pattern: str, schema: str = None):
        """
        :param dataset: table name for this stream
        :param provider: provider specific mapping as described in spec.json
        :param format: file format specific mapping as described in spec.json
        :param path_pattern: glob-style pattern for file-matching (https://facelessuser.github.io/wcmatch/glob/)
        :param schema: JSON-syntax user provided schema, defaults to None
        """
        self.dataset = dataset
        self._path_pattern = path_pattern
        self._provider = provider
        self._format = format
        self._schema: Dict[str, Any] = {}
        if schema:
            self._schema = self._parse_user_input_schema(schema)
        self.master_schema: Dict[str, Any] = None
        LOGGER.info(f"initialised stream with format: {format}")

    def _parse_user_input_schema(self, schema: str) -> Dict[str, Any]:
        """
        If the user provided a schema, we run this method to convert to a python dict and verify it
        This verifies:
            - that the provided string is valid JSON
            - that it is a key:value map with no nested values (objects or arrays)
            - that all values in the map correspond to a JsonSchema datatype
        If this passes, we are confident that the user-provided schema is valid and will work as expected with the rest of the code

        :param schema: JSON-syntax user provided schema
        :raises ConfigurationError: if any of the verification steps above fail
        :return: the input schema (json string) as a python dict
        """
        try:
            py_schema: Dict[str, Any] = json.loads(schema)
        except json.decoder.JSONDecodeError as err:
            error_msg = f"Failed to parse schema {repr(err)}\n{schema}\n{format_exc()}"
            raise ConfigurationError(error_msg) from err
        # enforce all keys and values are of type string as required (i.e. no nesting)
        if not all(isinstance(k, str) and isinstance(v, str) for k, v in py_schema.items()):
            raise ConfigurationError("Invalid schema provided, all column names and datatypes must be in string format")
        # enforce all values (datatypes) are valid JsonSchema datatypes
        if any(datatype not in JSON_TYPES for datatype in py_schema.values()):
            raise ConfigurationError(f"Invalid schema provided, datatypes must each be one of {JSON_TYPES}")

        return py_schema

    @property
    def name(self) -> str:
        return self.dataset

    @property
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        return None

    @property
    def fileformatparser_class(self) -> type:
        """
        :return: reference to the relevant fileformatparser class e.g. CsvParser
        """
        filetype = self._format.get("filetype")
        file_reader = FORMAT_PARSER_MAP.get(filetype)
        if not file_reader:
            raise RuntimeError(f"Detected mismatched file format '{filetype}'. Available values: '{list(FORMAT_PARSER_MAP.keys())}''.")
        return file_reader

    @property
    @abstractmethod
    def storagefile_class(self) -> type:
        """
        Override this to point to the relevant provider-specific StorageFile class e.g. S3File

        :return: reference to relevant class
        """

    @abstractmethod
    def file_iterator(self) -> Iterator[FileInfo]:
        """
        Provider-specific method to iterate through bucket/container/etc. and yield a FileInfo for each file.
        This supplies the 'FileInfo' to use in StorageFile(), which has all file properties (last_modified, key, size).
        All these meta options are saved during loading of files' list.

        :yield: FileInfo object to use in StorageFile()
        """

    def pattern_matched_file_iterator(self, file_infos: Iterable[FileInfo]) -> Iterator[FileInfo]:
        """
        iterates through file_infos and yields only those file_infos that match user-provided path patterns

        :param file_infos: typically the output of file_iterator()
        :yield: FileInfo object to use in StorageFile(), if matching on user-provided path patterns
        """
        for file_info in file_infos:
            if globmatch(file_info.key, self._path_pattern, flags=GLOBSTAR | SPLIT):
                yield file_info

    @lru_cache(maxsize=None)
    def get_time_ordered_file_infos(self) -> List[FileInfo]:
        """
        Iterates through pattern_matched_file_iterator(), acquiring FileInfo objects
        with last_modified property of each file to return in time ascending order.
        Caches results after first run of method to avoid repeating network calls as this is used more than once

        :return: list in time-ascending order
        """
        return sorted(self.pattern_matched_file_iterator(self.file_iterator()), key=lambda file_info: file_info.last_modified)

    def _get_schema_map(self) -> Mapping[str, Any]:
        return_schema: Dict[str, Any] = None
        if self._schema != {}:
            return_schema = deepcopy(self._schema)
        else:  # we have no provided schema or schema state from a previous incremental run
            return_schema = self._get_master_schema()

        return_schema[self.ab_additional_col] = "object"
        return_schema[self.ab_last_mod_col] = "string"
        return_schema[self.ab_file_name_col] = "string"
        return return_schema

    def get_json_schema(self) -> Mapping[str, Any]:
        """
        :return: the JSON schema representing this stream.
        """
        # note: making every non-airbyte column nullable for compatibility
        properties: Mapping[str, Any] = {
            column: {"type": ["null", typ]} if column not in self.airbyte_columns else {"type": typ}
            for column, typ in self._get_schema_map().items()
        }

        properties[self.ab_last_mod_col]["format"] = "date-time"
        return {"type": "object", "properties": properties}

    def _get_master_schema(self, min_datetime: datetime = None) -> Dict[str, Any]:
        """
        In order to auto-infer a schema across many files and/or allow for additional properties (columns),
            we need to determine the superset of schemas across all relevant files.
        This method iterates through get_time_ordered_file_infos() obtaining the inferred schema (process implemented per file format),
            to build up this superset schema (master_schema).
        This runs datatype checks to Warn or Error if we find incompatible schemas (e.g. same column is 'date' in one file but 'float' in another).
        This caches the master_schema after first run in order to avoid repeated compute and network calls to infer schema on all files.

        :param min_datetime: if passed, will only use files with last_modified >= this to determine master schema

        :raises RuntimeError: if we find datatype mismatches between files or between a file and schema state (provided or from previous inc. batch)
        :return: A dict of the JSON schema representing this stream.
        """
        # TODO: could implement a (user-beware) 'lazy' mode that skips schema checking to improve performance
        # TODO: could utilise min_datetime to add a start_date parameter in spec for user
        if self.master_schema is None:
            master_schema = deepcopy(self._schema)

            file_reader = self.fileformatparser_class(self._format)

            for file_info in self.get_time_ordered_file_infos():
                # skip this file if it's earlier than min_datetime
                if (min_datetime is not None) and (file_info.last_modified < min_datetime):
                    continue

                storagefile = self.storagefile_class(file_info, self._provider)
                with storagefile.open(file_reader.is_binary) as f:
                    this_schema = file_reader.get_inferred_schema(f)

                if this_schema == master_schema:
                    continue  # exact schema match so go to next file

                # creates a superset of columns retaining order of master_schema with any additional columns added to end
                column_superset = list(master_schema.keys()) + [c for c in this_schema.keys() if c not in master_schema.keys()]
                # this compares datatype of every column that the two schemas have in common
                for col in column_superset:
                    if (col in master_schema.keys()) and (col in this_schema.keys()) and (master_schema[col] != this_schema[col]):
                        # If this column exists in a provided schema or schema state, we'll WARN here rather than throw an error
                        # this is to allow more leniency as we may be able to coerce this datatype mismatch on read according to
                        # provided schema state. If not, then the read will error anyway
                        if col in self._schema.keys():
                            LOGGER.warn(
                                f"Detected mismatched datatype on column '{col}', in file '{storagefile.url}'. "
                                + f"Should be '{master_schema[col]}', but found '{this_schema[col]}'. "
                                + f"Airbyte will attempt to coerce this to {master_schema[col]} on read."
                            )
                        # else we're inferring the schema (or at least this column) from scratch and therefore
                        # throw an error on mismatching datatypes
                        else:
                            raise RuntimeError(
                                f"Detected mismatched datatype on column '{col}', in file '{storagefile.url}'. "
                                + f"Should be '{master_schema[col]}', but found '{this_schema[col]}'."
                            )

                # missing columns in this_schema doesn't affect our master_schema, so we don't check for it here

                # add to master_schema any columns from this_schema that aren't already present
                for col, datatype in this_schema.items():
                    if col not in master_schema.keys():
                        master_schema[col] = datatype

            LOGGER.info(f"determined master schema: {master_schema}")
            self.master_schema = master_schema

        return self.master_schema

    def stream_slices(
        self, sync_mode: SyncMode, cursor_field: List[str] = None, stream_state: Mapping[str, Any] = None
    ) -> Iterable[Optional[Dict[str, Any]]]:
        """
        This builds full-refresh stream_slices regardless of sync_mode param.
        For full refresh, 1 file == 1 stream_slice.
        The structure of a stream slice is [ {file}, ... ].
        In incremental mode, a stream slice may have more than one file so we mirror that format here.
        Incremental stream_slices are implemented in the IncrementalFileStream child class.
        """

        # TODO: this could be optimised via concurrent reads, however we'd lose chronology and need to deal with knock-ons of that
        # we could do this concurrently both full and incremental by running batches in parallel
        # and then incrementing the cursor per each complete batch
        for file_info in self.get_time_ordered_file_infos():
            yield {"files": [{"storage_file": self.storagefile_class(file_info, self._provider)}]}

    def _match_target_schema(self, record: Dict[str, Any], target_columns: List) -> Dict[str, Any]:
        """
        This method handles missing or additional fields in each record, according to the provided target_columns.
        All missing fields are added, with a value of None (null)
        All additional fields are packed into the _ab_additional_properties object column
        We start off with a check to see if we're already lined up to target in order to avoid unnecessary iterations (useful if many columns)

        :param record: json-like representation of a data row {column:value}
        :param target_columns: list of column names to mutate this record into (obtained via self._get_schema_map().keys() as of now)
        :return: mutated record with columns lining up to target_columns
        """
        compare_columns = [c for c in target_columns if c not in [self.ab_last_mod_col, self.ab_file_name_col]]
        # check if we're already matching to avoid unnecessary iteration
        if set(list(record.keys()) + [self.ab_additional_col]) == set(compare_columns):
            record[self.ab_additional_col] = {}
            return record
        # missing columns
        for c in [col for col in compare_columns if col != self.ab_additional_col]:
            if c not in record.keys():
                record[c] = None
        # additional columns
        record[self.ab_additional_col] = {c: deepcopy(record[c]) for c in record.keys() if c not in compare_columns}
        for c in record[self.ab_additional_col]:
            del record[c]

        return record

    def _add_extra_fields_from_map(self, record: Dict[str, Any], extra_map: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Simple method to take a mapping of columns:values and add them to the provided record

        :param record: json-like representation of a data row {column:value}
        :param extra_map: map of additional columns and values to add
        :return: mutated record with additional fields
        """
        for key, value in extra_map.items():
            record[key] = value
        return record

    def _read_from_slice(
        self,
        file_reader: AbstractFileParser,
        stream_slice: Mapping[str, Any],
        stream_state: Mapping[str, Any] = None,
    ) -> Iterable[Mapping[str, Any]]:
        """
        Uses provider-relevant StorageFile to open file and then iterates through stream_records() using format-relevant AbstractFileParser.
        Records are mutated on the fly using _match_target_schema() and _add_extra_fields_from_map() to achieve desired final schema.
        Since this is called per stream_slice, this method works for both full_refresh and incremental.
        """
        for file_item in stream_slice["files"]:
            storage_file: StorageFile = file_item["storage_file"]
            with storage_file.open(file_reader.is_binary) as f:
                # TODO: make this more efficient than mutating every record one-by-one as they stream
                for record in file_reader.stream_records(f):
                    schema_matched_record = self._match_target_schema(record, list(self._get_schema_map().keys()))
                    complete_record = self._add_extra_fields_from_map(
                        schema_matched_record,
                        {
                            self.ab_last_mod_col: datetime.strftime(storage_file.last_modified, self.datetime_format_string),
                            self.ab_file_name_col: storage_file.url,
                        },
                    )
                    yield complete_record
        LOGGER.info("finished reading a stream slice")

    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: List[str] = None,
        stream_slice: Mapping[str, Any] = None,
        stream_state: Mapping[str, Any] = None,
    ) -> Iterable[Mapping[str, Any]]:
        """
        The heavy lifting sits in _read_from_slice() which is full refresh / incremental agnostic
        """
        if stream_slice:
            file_reader = self.fileformatparser_class(self._format, self._get_master_schema())
            yield from self._read_from_slice(file_reader, stream_slice)


class IncrementalFilesStream(FilesStream, ABC):
    # TODO: ideally want to checkpoint after every file or stream slice rather than N records
    state_checkpoint_interval = None
    buffer_days = 3  # keeping track of all files synced in the last N days
    sync_all_files_always = False
    max_history_size = 1000000000

    @property
    def cursor_field(self) -> str:
        """
        :return: The name of the cursor field.
        """
        return self.ab_last_mod_col

    @staticmethod
    def file_in_history(file_info: FileInfo, history: dict) -> bool:
        for slot in history.values():
            if file_info.key in slot:
                return file_info.key in slot
        return False

    def _get_datetime_from_stream_state(self, stream_state: Mapping[str, Any] = None) -> datetime:
        """if no state, we default to 1970-01-01 in order to pick up all files present."""
        if stream_state is not None and self.cursor_field in stream_state.keys():
            return datetime.strptime(stream_state[self.cursor_field], self.datetime_format_string)
        else:
            return datetime.strptime("1970-01-01T00:00:00+0000", self.datetime_format_string)

    def get_updated_history(self, current_stream_state, latest_record_datetime, latest_record, current_parsed_datetime, state_date):
        """
        History is dict which basically groups files by their modified_at date.
        After reading each record we add its file to the history set if it wasn't already there.
        Then we drop from the history set any entries whose key is less than now - buffer_days
        """

        history = current_stream_state.get("history", {})

        file_modification_date = latest_record_datetime.strftime("%Y-%m-%d")

        # add record to history if record modified date in range delta start from state
        if latest_record_datetime.date() + timedelta(days=self.buffer_days) >= state_date:
            history_item = set(history.setdefault(file_modification_date, set()))
            history_item.add(latest_record[self.ab_file_name_col])
            history[file_modification_date] = history_item

        # reset history to new date state
        if current_parsed_datetime.date() != state_date:
            history = {
                date: history[date]
                for date in history
                if datetime.strptime(date, "%Y-%m-%d").date() + timedelta(days=self.buffer_days) >= state_date
            }

        return history

    def size_history_balancer(self, state_dict):
        """
        Delete history if state size limit reached
        """
        history = state_dict["history"]

        if history.__sizeof__() > self.max_history_size:
            self.sync_all_files_always = True
            state_dict.pop("history")

        return state_dict

    def get_updated_state(self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Inspects the latest record extracted from the data source and the current state object and return an updated state object.
        In the case where current_stream_state is null, we default to 1970-01-01 in order to pick up all files present.
        We also save the schema into the state here so that we can use it on future incremental batches, allowing for additional/missing columns.

        :param current_stream_state: The stream's current state object
        :param latest_record: The latest record extracted from the stream
        :return: An updated state object
        """
        state_dict: Dict[str, Any] = {}
        current_parsed_datetime = self._get_datetime_from_stream_state(current_stream_state)
        latest_record_datetime = datetime.strptime(
            latest_record.get(self.cursor_field, "1970-01-01T00:00:00+0000"), self.datetime_format_string
        )
        state_dict[self.cursor_field] = datetime.strftime(max(current_parsed_datetime, latest_record_datetime), self.datetime_format_string)

        state_dict["schema"] = self._get_schema_map()

        state_date = self._get_datetime_from_stream_state(state_dict).date()

        if not self.sync_all_files_always:
            state_dict["history"] = self.get_updated_history(
                current_stream_state, latest_record_datetime, latest_record, current_parsed_datetime, state_date
            )

        return self.size_history_balancer(state_dict)

    def need_to_skip_file(self, stream_state, file_info):
        """
        Skip this file if last_mod is earlier than our cursor value from state and already in history
        or skip this file if last_mod plus delta is earlier than our cursor value
        """
        file_in_history_and_last_modified_is_earlier_than_cursor_value = (
            stream_state is not None
            and self.cursor_field in stream_state.keys()
            and file_info.last_modified <= self._get_datetime_from_stream_state(stream_state)
            and self.file_in_history(file_info, stream_state.get("history", {}))
        )

        file_is_not_in_history_and_last_modified_plus_buffer_days_is_earlier_than_cursor_value = file_info.last_modified + timedelta(
            days=self.buffer_days
        ) < self._get_datetime_from_stream_state(stream_state) and not self.file_in_history(file_info, stream_state.get("history", {}))

        return (
            file_in_history_and_last_modified_is_earlier_than_cursor_value
            or file_is_not_in_history_and_last_modified_plus_buffer_days_is_earlier_than_cursor_value
        )

    def stream_slices(
        self, sync_mode: SyncMode, cursor_field: List[str] = None, stream_state: Mapping[str, Any] = None
    ) -> Iterable[Optional[Dict[str, Any]]]:
        """
        Builds either full_refresh or incremental stream_slices based on sync_mode.
        An incremental stream_slice is a group of all files with the exact same last_modified timestamp.
        This ensures we only update the cursor state to a given timestamp after ALL files with that timestamp have been successfully read.

        Slight nuance: as we iterate through get_time_ordered_file_infos(),
        we yield the stream_slice containing file(s) up to and Excluding the file on the current iteration.
        The stream_slice is then cleared (if we yielded it) and this iteration's file appended to the (next) stream_slice
        """
        if sync_mode == SyncMode.full_refresh:
            yield from super().stream_slices(sync_mode=sync_mode, cursor_field=cursor_field, stream_state=stream_state)

        else:
            # if necessary and present, let's update this object's schema attribute to the schema stored in state
            # TODO: ideally we could do this on __init__ but I'm not sure that's possible without breaking from cdk style implementation
            if self._schema == {} and stream_state is not None and "schema" in stream_state.keys():
                self._schema = stream_state["schema"]

            # logic here is to bundle all files with exact same last modified timestamp together in each slice
            prev_file_last_mod: datetime = None  # init variable to hold previous iterations last modified
            grouped_files_by_time: List[Dict[str, Any]] = []
            for file_info in self.get_time_ordered_file_infos():
                if self.need_to_skip_file(stream_state, file_info):
                    continue

                # check if this file belongs in the next slice, if so yield the current slice before this file
                if (prev_file_last_mod is not None) and (file_info.last_modified != prev_file_last_mod):
                    yield {"files": grouped_files_by_time}
                    grouped_files_by_time.clear()

                # now we either have an empty stream_slice or a stream_slice that this file shares a last modified with, so append it
                grouped_files_by_time.append({"storage_file": self.storagefile_class(file_info, self._provider)})
                # update our prev_file_last_mod to the current one for next iteration
                prev_file_last_mod = file_info.last_modified

            # now yield the final stream_slice. This is required because our loop only yields the slice previous to its current iteration.
            if len(grouped_files_by_time) > 0:
                yield {"files": grouped_files_by_time}
            else:
                # in case we have no files
                yield None

    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: List[str] = None,
        stream_slice: Mapping[str, Any] = None,
        stream_state: Mapping[str, Any] = None,
    ) -> Iterable[Mapping[str, Any]]:
        """
        The heavy lifting sits in _read_from_slice() which is full refresh / incremental agnostic.
        We override this for incremental so we can pass our minimum datetime from state into _get_master_schema().
        This means we only parse the schema of new files on incremental runs rather than all files in the bucket.
        """
        if stream_slice:
            if sync_mode == SyncMode.full_refresh:
                yield from super().read_records(sync_mode, cursor_field, stream_slice, stream_state)
            else:
                file_reader = self.fileformatparser_class(
                    self._format, self._get_master_schema(self._get_datetime_from_stream_state(stream_state))
                )
                yield from self._read_from_slice(file_reader, stream_slice)
